"""Define several typing for convenient use"""

from typing import Union, Callable, Optional, Any, List, Tuple

import numpy as np
from ase import Atoms
from pymatgen.core import Structure, Molecule

OptStrOrCallable = Optional[Union[str, Callable[..., Any]]]
StructureOrMolecule = Union[Structure, Molecule]
StructureOrMoleculeOrAtoms = Union[Structure, Molecule, Atoms]
VectorLike = Union[List[float], np.ndarray]
ListTuple = Union[List, Tuple]



# -*- coding: utf-8 -*-

# @Time   : 2019/5/23 20:40
# @Author : Administrator
# @Project : feature_preparation
# @FileName: backforward.py
# @Software: PyCharm

"""Forward_and_back feature elimination for feature ranking"""

import copy
import warnings
from functools import partial
from typing import List

import numpy as np
from mgetool.tool import parallelize
from sklearn.base import BaseEstimator
from sklearn.base import MetaEstimatorMixin
from sklearn.base import clone
from sklearn.feature_selection import SelectorMixin
from sklearn.metrics import check_scoring
from sklearn.model_selection._validation import _score, cross_val_score
from sklearn.utils.metaestimators import if_delegate_has_method, _safe_split
from sklearn.utils.validation import check_is_fitted, check_X_y, check_random_state

from featurebox.selection.mutibase import MutiBase


def _baf_single_fit(train, test, baf, estimator, X, y, scorer, random_state):
    """"""
    X_train, y_train = _safe_split(estimator, X, y, train)
    X_test, y_test = _safe_split(estimator, X, y, test, train)
    baf_i = clone(baf)
    baf_i.random_state = random_state
    baf_i.refit = True
    baf_i._fit(X_train, y_train)
    return baf_i.support_, _score(baf_i.estimator_, baf_i.transform(X_test, ), y_test, scorer), baf_i.score_


class BackForward(BaseEstimator, MetaEstimatorMixin, SelectorMixin, MutiBase):
    """
    BackForward method to selected features.

    estimator:
        A supervised sklearn learning estimator with a ``_fit`` method that provides
        information about feature importance either through a ``coef_``
        attribute or through a ``feature_importances_`` attribute.

    n_type_feature_to_select: int or None (default=None)
        The number of feature to selection. If `None`, selection the features with best score.

    n_feature_: int
        The number of selected feature.

    support_: array of shape [n_feature]
        The mask of selected feature.

    estimator_: object
        The external estimator _fit on the reduced dataset.

    Examples
    --------

    >>> from sklearn.datasets import load_boston
    >>> from sklearn.svm import SVR
    >>> X,y = load_boston(return_X_y=True)
    >>> svr= SVR()
    >>> bf = BackForward(svr,primary_feature=4, random_state=1)
    >>> new_x = bf.fit_transform(X,y)
    >>> bf.support_
    array([False, False, False, False, False, False, False, False, False,
           False,  True, False,  True])

    Examples
    --------
    if score and predict is used, the refit could be set True and make sure the data is spilted, due to the refit
     used all data in fit() function.

    >>> from sklearn.datasets import load_boston
    >>> from sklearn.svm import SVR
    >>> X,y = load_boston(return_X_y=True)
    >>> svr= SVR()
    >>> bf = BackForward(svr,primary_feature=4, random_state=1, refit=True)
    >>> new_x = bf.fit_transform(X[:50],y[:50])
    >>> test_score = bf.score(X[50:],y[50:])
    ...

    Examples
    --------
    if GridSearchCV, the refit could be set True and return the cv score.

    >>> from sklearn.datasets import load_boston
    >>> from sklearn.svm import SVR
    >>> from sklearn import model_selection
    >>> X,y = load_boston(return_X_y=True)
    >>> svr= SVR()
    >>> gd = model_selection.GridSearchCV(svr,param_grid={"C":[1,10]},n_jobs=1,cv=5)
    >>> bf = BackForward(gd,primary_feature=4, random_state=1, refit=True)
    >>> new_x = bf.fit_transform(X,y)
    >>> test_score = bf.score(X,y)
    ...
    """

    def __init__(self, estimator: BaseEstimator, n_type_feature_to_select: int = None, primary_feature: int = None,
                 muti_grade: int = 2, muti_index: List = None, refit=False,
                 must_index: List = None, tolerant: float = 0.01, verbose: int = 0, random_state: int = None):
        """

        Parameters
        ----------
        estimator:sklearn.estimator
            sklearn.estimator
        n_type_feature_to_select:int
            force select number max
        primary_feature:int
             primary features to start loop, default n_features//2.
        muti_grade: int
            group number
        muti_index:
            group index
        must_index:
            must selection index
        tolerant:
            tolerant for rank compare.
        verbose:int
            print or not
        random_state:int
            random_state
        refit:refit
            refit or not. if refit, the model would use all data.
        """
        super().__init__(muti_grade=muti_grade, muti_index=muti_index, must_index=must_index)
        self.estimator = estimator
        self.n_type_feature_to_select = n_type_feature_to_select
        self.primary_feature = primary_feature
        self.verbose = verbose
        self.score_ = []
        self.random_state = random_state
        self.tolerant = tolerant
        self.refit = refit

    @property
    def _estimator_type(self):
        return self.estimator._estimator_type

    def fit(self, X, y):
        """Fit the baf model and then the underlying estimator on the selected
           feature.

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape = [n_samples, n_feature]
            The training input0 samples.

        y : array-like, shape = [n_samples]
            The target values.
        """
        return self._fit(X, y)

    def _fit(self, x, y):

        estimator = clone(self.estimator)

        def add_slice(slice10, slice20):
            best0 = score(slices=slice10)

            for add1 in list(slice20):
                # or random.choice() >= score(slice1):
                slice10.append(add1)
                test = score(slices=slice10)
                slice1_, test_ = sub_slice(slice10)
                if test_ >= max(best0, test):
                    best0 = test_
                    slice10 = slice1_
                elif test > max(best0, test_) + 2 * self.tolerant:
                    best0 = test
                else:
                    slice10.remove(add1)
                if self.verbose > 0:
                    print("Fitting estimator with {} feature {}".format(len(slice10), best0))
                self.score_.append((tuple(slice10), best0))
            return slice10

        def sub_slice(slice10):
            best0 = score(slices=slice10)
            slice0 = list(copy.deepcopy(slice10))
            slice10 = list(slice10)
            ran.shuffle(slice10)
            if self.check_must:
                slice10 = [_ for _ in slice10 if _ not in self.check_must]
            for sub in list(slice10):
                slice0.remove(sub)
                test0 = score(slices=slice0)
                if test0 > best0 - self.tolerant:
                    best0 = test0
                else:
                    slice0.append(sub)
                # print(slice0, best)
            return slice0, best0

        def sub_slice_force(slice10):
            best0 = score(slices=slice10)
            slice0 = list(copy.deepcopy(slice10))
            ran.shuffle(slice0)

            while len(slice0) > self.n_type_feature_to_select:
                slice_all = []
                slice10 = list(copy.deepcopy(slice0))
                if self.check_must:
                    slice10 = [_ for _ in slice10 if _ not in self.check_must]
                le = len(slice10)
                while le:
                    slice_all.append(copy.deepcopy(slice0))
                    le -= 1
                for slice_, sub in zip(slice_all, slice10):
                    slice_.remove(sub)
                test0 = [score(slices=i) for i in slice_all]
                index = int(np.argmax(test0))
                best0 = max(test0)
                slice0 = slice_all[index]
                # print(slice0, best0)
                if self.verbose > 0:
                    print("Fitting estimator with {} feature {}".format(len(slice0), best0))
                self.score_.append((tuple(slice0), best0))
            return slice0, best0

        def score_pri(slices, x0, y0):
            slices = list(slices)
            if len(slices) <= 1:
                score0 = - np.inf
            else:
                slices = self.feature_unfold(slices)
                data_x0 = x0[:, slices]

                estimator.fit(data_x0, y0)
                if hasattr(estimator, 'best_score_'):
                    score0 = np.mean(estimator.best_score_)
                else:
                    score0 = np.mean(cross_val_score(estimator, data_x0, y0, cv=5))
            return score0

        score = partial(score_pri, x0=x, y0=y)

        self.score_ = []
        x, y = check_X_y(x, y, "csc")
        assert all((self.check_must, self.check_muti)) in [True, False]
        # Initialization
        if self.check_muti:
            n_feature = (self.muti_index[1] - self.muti_index[0]) // self.muti_grade + self.muti_index[0]
        else:
            n_feature = x.shape[1]
        if self.primary_feature is None:
            primary_feature = n_feature // 2
        else:
            primary_feature = self.primary_feature

        feature_list = list(range(x.shape[1]))
        fold_feature_list = self.feature_fold(feature_list)

        ran = check_random_state(self.random_state)
        slice1 = ran.choice(fold_feature_list, primary_feature, replace=False)
        slice1 = list(self.feature_must_fold(slice1))
        slice2 = list(set(fold_feature_list) - set(slice1))
        ran.shuffle(slice2)

        slice1 = add_slice(slice1, slice2)
        if isinstance(self.n_type_feature_to_select, int) and len(slice1) > self.n_type_feature_to_select:
            slice1, best = sub_slice_force(slice1)

        select_feature = self.feature_unfold(slice1)
        su = np.zeros(x.shape[1], dtype=np.bool)
        su[select_feature] = 1
        self.support_ = su
        self.estimator_ = clone(self.estimator)
        if self.refit:
            if not hasattr(self.estimator_, 'best_score_'):
                warnings.warn(UserWarning(
                    "The self.estimator_ :{} used all the X,y data.".format(self.estimator_.__class__.__name__),
                    "please be careful with the later 'score' and 'predict'."))
            if hasattr(self.estimator_, 'best_score_') and hasattr(self.estimator_, "refit") \
                    and self.estimator_.refit is True:
                warnings.warn(UserWarning(
                    "The self.estimator_ :{} used all the X,y data.".format(self.estimator_.__class__.__name__),
                    "please be careful with the later 'score' and 'predict'."))
            self.estimator_.fit(x[:, select_feature], y)
        self.n_feature_ = len(select_feature)

        return self

    @if_delegate_has_method(delegate='estimator_')
    def predict(self, X):
        """Reduce X to the selected feature and then using the underlying estimator to predict.
         Only available ``refit=True``.

        Parameters
        ----------
        X : array of shape [n_samples, n_feature]
            The input0 samples.

        Returns
        -------
        y : array of shape [n_samples]
            The predicted target values.
        """
        check_is_fitted(self, 'estimator_')
        return self.estimator_.predict(self.transform(X))

    @if_delegate_has_method(delegate='estimator_')
    def score(self, X, y):
        """Reduce X to the selected feature and then return the score of the underlying estimator.
        Only available ``refit=True``.

        Parameters
        ----------
        X : array of shape [n_samples, n_feature]
            The input0 samples.

        y : array of shape [n_samples]
            The target values.
        """
        check_is_fitted(self, 'estimator_')
        return self.estimator_.score(self.transform(X), y)

    def _get_support_mask(self):
        check_is_fitted(self, 'support_')
        return self.support_


def _multi_time_fit(random_state, baf, X, y, scorer):
    """"""
    baf_i = clone(baf)
    baf_i.random_state = random_state
    baf_i.refit = True
    baf_i._fit(X, y)
    return baf_i.support_, _score(baf_i.estimator_, baf_i.transform(X, ), y, scorer), baf_i.score_


class BackForwardStable(MetaEstimatorMixin, SelectorMixin, BaseEstimator):
    """
    BackForwardStable.
    Run with different order for more Stable (Not guaranteed).

    estimator : object
        A supervised learning estimator with a ``_fit`` method that provides
        information about feature importance either through a ``coef_``
        attribute or through a ``feature_importances_`` attribute.

    scoring : string, callable or None, optional, (default=None)
        A string (see model evaluation documentation) or
        a scorer callable object / function with signature
        ``scorer(estimator, X, y)``.

    verbose : int, (default=0)
        Controls verbosity of output.

    n_jobs : int or None, optional (default=None)
        Number of cores to run in parallel while fitting across folds.
        ``None`` means 1 unless in a :obj:`joblib.parallel_backend` context.
        ``-1`` means using all processors. See :term:`Glossary <n_jobs>`
        for more details.

    n_feature_ : int
        The number of selected feature with cross-validation.

    support_ : array of shape [n_feature]
        The mask of selected feature.

    estimator_ : object
        The external estimator _fit on the reduced dataset.

    n_jobs : int or None
        Number of cores to run in parallel while fitting across folds.
        ``None`` means 1 and ``-1`` means using all processors.


    Examples
    --------

    >>> from sklearn.datasets import load_boston
    >>> from sklearn.svm import SVR
    >>> X,y = load_boston(return_X_y=True)
    >>> svr= SVR()
    >>> bf = BackForwardStable(svr,primary_feature=9,  random_state=1)
    >>> new_x = bf.fit_transform(X,y)
    >>> bf.support_
    array([False, False, False, False, False, False, False, False, False,
           False,  True, False,  True])
    >>> bf.score_
    0.642826382278419

    Examples
    --------

    if score and predict is used, the refit could be set True and make sure the data is spilted, due to the refit
    used all data in fit() function.

    >>> from sklearn.datasets import load_boston
    >>> from sklearn.svm import SVR
    >>> X,y = load_boston(return_X_y=True)
    >>> svr= SVR()
    >>> bf = BackForwardStable(svr,primary_feature=4, random_state=1, refit=True)
    >>> new_x = bf.fit_transform(X[:50],y[:50])
    >>> test_score = bf.score(X[50:],y[50:])
    ...

    Examples
    --------

    if GridSearchCV, the refit could be set True and return the cv score.

    >>> from sklearn.datasets import load_boston
    >>> from sklearn.svm import SVR
    >>> from sklearn import model_selection
    >>> X,y = load_boston(return_X_y=True)
    >>> svr= SVR()
    >>> gd = model_selection.GridSearchCV(svr,param_grid={"C":[1,10]},n_jobs=1,cv=5)
    >>> bf = BackForward(gd,primary_feature=4, random_state=1, refit=True)
    >>> new_x = bf.fit_transform(X,y)
    >>> test_score = bf.score(X,y)
    ...
    """

    def __init__(self, estimator: BaseEstimator, n_type_feature_to_select: int = None,
                 primary_feature: int = None, muti_grade: int = 2, muti_index: List = None,
                 must_index: List = None, verbose: int = 0, random_state: int = None,
                 times: int = 5, scoring: str = "r2", n_jobs: int = None, refit=False):
        """

        Parameters
        ----------
        estimator:sklearn.estimator
            sklearn.estimator
        n_type_feature_to_select:int
            force select number max
        primary_feature:int
             expectation select number
        muti_grade: int
            group number
        muti_index: list
            group index
        must_index: list
            must selection index
        random_state:int
            random_state
        scoring : string, callable or None, optional, (default=None)
            A string (see model evaluation documentation) or
            a scorer callable object / function with signature
            ``scorer(estimator, X, y)``.
        verbose : int, (default=0)
            Controls verbosity of output.
        n_jobs : int or None
            Number of cores to run in parallel while fitting across folds.
            ``None`` means 1 and ``-1`` means using all processors.
        refit:
            False, with refit, the model would used all data.
        """

        self.estimator = estimator
        self.times = times
        self.scoring = scoring
        self.verbose = verbose
        self.n_jobs = n_jobs
        self.n_type_feature_to_select = n_type_feature_to_select
        self.primary_feature = primary_feature
        self.muti_grade = muti_grade
        self.muti_index = muti_index
        self.must_index = must_index
        self.score_ = []
        self.random_state = random_state
        self.refit = refit

    def fit(self, X, y, groups=None):
        """Fit the baf model and automatically tune the number of selected feature.

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape = [n_samples, n_feature]
            Training vector, where `n_samples` is the number of samples and
            `n_feature` is the total number of feature.

        y : array-like, shape = [n_samples]
            Target values (integers for classification, real numbers for
            regression).

        groups : array-like, shape = [n_samples], optional
            cal_group labels for the samples used while splitting the dataset into
            train/test set.
        """
        X, y = check_X_y(X, y, "csr")
        # Initialization
        estimator = clone(self.estimator)
        scorer = check_scoring(estimator, scoring=self.scoring)
        ran = check_random_state(self.random_state)

        baf = BackForward(estimator=estimator,
                          n_type_feature_to_select=self.n_type_feature_to_select,
                          verbose=self.verbose, primary_feature=self.primary_feature,
                          muti_grade=self.muti_grade, muti_index=self.muti_index,
                          must_index=self.must_index, random_state=ran)
        rans = ran.randint(0, 1000, self.times)

        func = partial(_multi_time_fit, baf=baf, X=X, y=y, scorer=scorer)

        scores = parallelize(n_jobs=self.n_jobs, func=func, iterable=rans, respective=False)

        support, scores, score_step = zip(*scores)
        best_support = support[np.argmax(scores)]
        best_score = max(scores)
        # Re-execute an elimination with best_k over the whole set

        # Set final attributes
        self.support_step = score_step
        self.support_ = best_support
        self.score_ = best_score
        self.estimator_ = clone(self.estimator)
        if self.refit:
            if not hasattr(self.estimator_, 'best_score_'):
                warnings.warn(UserWarning(
                    "The self.estimator_ :{} used all the X,y data.".format(self.estimator_.__class__.__name__),
                    "please be careful with the later 'score' and 'predict'."))
            if hasattr(self.estimator_, 'best_score_') and hasattr(self.estimator_, "refit") \
                    and self.estimator_.refit is True:
                warnings.warn(UserWarning(
                    "The self.estimator_ :{} used all the X,y data.".format(self.estimator_.__class__.__name__),
                    "please be careful with the later 'score' and 'predict'."))
            self.estimator_.fit(X[:, self.support_], y)
        self.n_feature_ = np.count_nonzero(support)
        return self

    @if_delegate_has_method(delegate='estimator_')
    def predict(self, X):
        """Reduce X to the selected feature and then Fit using the underlying estimator, only with refit.
        Only available ``refit=True``.

        Parameters
        ----------
        X : array of shape [n_samples, n_feature]
            The input0 samples.

        Returns
        -------
        y : array of shape [n_samples]
            The predicted target values.
        """
        check_is_fitted(self, 'estimator_')
        return self.estimator_.predict(self.transform(X))

    @if_delegate_has_method(delegate='estimator_')
    def score(self, X, y):
        """Reduce X to the selected feature and then return the score of the underlying estimator, only with refit.
        Only available ``refit=True``.

        Parameters
        ----------
        X : array of shape [n_samples, n_feature]
            The input0 samples.

        y : array of shape [n_samples]
            The target values.
        """
        check_is_fitted(self, 'estimator_')
        return self.estimator_.score(self.transform(X), y)

    def _get_support_mask(self):
        check_is_fitted(self, 'support_')
        return self.support_

# from sklearn.datasets import load_boston
# from sklearn.svm import SVR
# X,y = load_boston(return_X_y=True)
# svr= SVR()
# bf = BackForwardCV(svr, primary_feature=9,  random_state=1)
# new_x = bf.fit_transform(X,y)
# bf.support_



import unittest

import pandas as pd

from featurebox.featurizers.state.state_mapper import StructurePymatgenPropMap


class MyTestCase(unittest.TestCase):
    def setUp(self) -> None:
        self.data = pd.read_pickle("data_structure.pkl_pd")
        self.data0 = self.data[0]
        self.data0_3 = self.data[:3]

    # def test_something(self):
    #     sppm = StructurePymatgenPropMap()
    #     data = sppm.fit_transform(self.data0_3)
    #     data = sppm.fit_transform(self.data0_3)

    def test_something2(self):
        sppm = StructurePymatgenPropMap(prop_name=["lengths", "angles", "get_brillouin_zone()"])
        data = sppm.fit_transform([i._lattice for i in self.data0_3])

    def test_something3(self):
        sppm = StructurePymatgenPropMap(
            prop_name=["_lattice.lengths", "_lattice.angles", "_lattice.get_brillouin_zone()"])
        data = sppm.fit_transform(self.data0_3)
        data = sppm.fit_transform(self.data0_3)


if __name__ == '__main__':
    unittest.main()



import unittest

import numpy as np

from featurebox.featurizers.base_transform import BaseFeature, DummyConverter


class MyTestCase(unittest.TestCase):
    def test_json(self):
        bf = BaseFeature()
        bf.to_json()
        d = bf.as_dict()
        bf.from_dict(d)

    def test_transofrm(self):
        bf = BaseFeature(n_jobs=1)
        x1 = [1, 2, 3, 4, 5, 6]
        x2 = [1, 2, 3, 4, 5, 6.0]
        x3 = np.array([[1, 2, 3, 4, 5, 6.0], [1, 2, 3, 4, 5, 6.0]])
        x4 = np.array([[1, 2, 3, 4, 5, 6.0]])
        x5 = np.array([1, 2, 3, 4, 5, 6.0])
        newx = bf.fit_transform(x1)
        print(newx)
        newx = bf.fit_transform(x2)
        print(newx)
        newx = bf.fit_transform(x3)
        print(newx)
        newx = bf.fit_transform(x4)
        print(newx)
        newx = bf.fit_transform(x5)
        print(newx)

    def test_transofrm2(self):
        bf = DummyConverter(n_jobs=1)
        x1 = [1, 2, 3, 4, 5, 6]
        x2 = [1, 2, 3, 4, 5, 6.0]
        x3 = np.array([[1, 2, 3, 4, 5, 6.0], [1, 2, 3, 4, 5, 6.0]])
        x4 = np.array([[1, 2, 3, 4, 5, 6.0]])
        x5 = np.array([1, 2, 3, 4, 5, 6.0])
        newx = bf.fit_transform(x1)
        print(newx)
        newx = bf.fit_transform(x2)
        print(newx)
        newx = bf.fit_transform(x3)
        print(newx)
        newx = bf.fit_transform(x4)
        print(newx)
        newx = bf.fit_transform(x5)
        print(newx)

    def test_fea(self):
        bf = DummyConverter(n_jobs=1)
        bf.set_feature_labels(["name"])
        print(bf.feature_labels)


if __name__ == '__main__':
    unittest.main()



# Numpy Version

from itertools import combinations, combinations_with_replacement

import numpy as np
from ase.neighborlist import NeighborList


class wACSF:
    """A class for calculating weighted atom-centered symmetry functions (wACSFs).
    
    The forms of the functions are consistent with the functions presented in:
    Gastegger, M., et. al. (2018). The Journal of chemical physics, 148(24), 241709.

    Note that this code does not implement Rs in the angular wACSF.
        
    Parameters
    ----------
    symmetry_parameters: dict
        The user-defined parameters for the symmetry functions calculations.
        i.e. {'G2': {'eta': [0.1, 0.2,], 'Rs': [0.0, 0.1]}}
    Rc: float
        The symmetry functions will be calculated within this radius.
    derivative: bool
        If True, calculate the derivatives.
    stress: bool
        If True, calculate the virial stress contribution.
    """

    def __init__(self, symmetry_parameters=None,
                 Rc=6.5, derivative=True, stress=False):
        if symmetry_parameters is None:
            symmetry_parameters = {'G2': {'eta': [0.1, 0.2, ], 'Rs': [0.0, 0.1]}}
        self._type = 'wACSF'

        # Set up the symmetry parameters keywords. If a string are not in the
        # keyword, code will return an error.
        G2_keywords = ['Rs', 'eta']
        G4_keywords = ['Rs', 'eta', 'lambda', 'zeta']
        G5_keywords = G4_keywords

        # Setting up parameters for each of the symmetry function type.
        self.G2_parameters = None
        self.G4_parameters = None
        self.G5_parameters = None

        for key, value in symmetry_parameters.items():
            if key == 'G2':
                self.G2_parameters = {'eta': [1], 'Rs': np.array([0.])}
                for key0, value0 in value.items():
                    if key0 in G2_keywords:
                        self.G2_parameters[key0] = to_array(value0)
                    else:
                        msg = f"{key0} is not available. " \
                              f"Choose from {G2_keywords}"
                        raise NotImplementedError(msg)

            elif key == 'G4':
                self.G4_parameters = {'Rs': np.array([0.]),
                                      'eta': [1], 'zeta': [1], 'lambda': [1]}
                for key0, value0 in value.items():
                    if key0 in G4_keywords:
                        self.G4_parameters[key0] = to_array(value0)
                    else:
                        msg = f"{key0} is not available. " \
                              f"Choose from {G4_keywords}"
                        raise NotImplementedError(msg)

            elif key == 'G5':
                self.G5_parameters = {'Rs': np.array([0.]),
                                      'eta': [1], 'zeta': [1], 'lambda': [1]}
                for key0, value0 in value.items():
                    if key0 in G5_keywords:
                        self.G5_parameters[key0] = to_array(value0)
                    else:
                        msg = f"{key0} is not available. " \
                              f"Choose from {G5_keywords}"
                        raise NotImplementedError(msg)

        self.Rc = Rc
        self.derivative = derivative
        self.stress = stress

    def calculate(self, crystal):
        """The symmetry functions are obtained through this `calculate` method.
        
        Parameters
        ----------
        crystal: object
            ASE Structure object.
                    
        Returns
        -------
        all_G: dict
            The user-defined symmetry functions that represent the crystal.
            Currently, there are 3 types of symmetry functions implemented. 
            Here are the order of the descriptors are printed out based on
            their symmetry parameters:
                - G2: ["Rs", "eta"]
                - G4: ["eta", "lambda", "zeta"]
                - G5: ["eta", "lambda", "zeta"]
        """
        self.crystal, self.total_atoms = crystal, len(crystal)
        atomic_numbers = np.array(crystal.get_atomic_numbers())
        vol = crystal.get_volume()

        # Initialize dict for the symmetry functions
        self.all_G = {'x': [], 'elements': []}
        if self.derivative:
            self.all_G['dxdr'] = []
        if self.stress:
            self.all_G['rdxdr'] = []

        # Obtain neighbors info.
        rc = [self.Rc / 2] * len(self.crystal)
        neighbors = NeighborList(rc, self_interaction=False, bothways=True, skin=0.0)
        neighbors.update(crystal)

        for i in range(len(crystal)):
            element = crystal.get_chemical_symbols()[i]
            indices, offsets = neighbors.get_neighbors(i)
            Z = []  # atomic numbers of neighbors

            assert len(indices) > 0, \
                f"There's no neighbor for this structure at Rc = {self.Rc} A."

            Ri = crystal.get_positions()[i]
            total_neighbors = len(indices)  # total number of neighbors of atom i

            Rj = np.zeros([total_neighbors, 3])
            IDs = np.zeros(total_neighbors, dtype=int)
            jks = np.array(list(combinations(range(total_neighbors), 2)))

            count = 0
            for j, offset in zip(indices, offsets):
                Rj[count, :] = crystal.positions[j] + np.dot(offset, crystal.get_cell())
                IDs[count] = j
                Z.append(crystal[j].number)
                count += 1
            Z = np.array(Z)

            Rij = Rj - Ri
            Dij = np.sqrt(np.sum(Rij ** 2, axis=1))

            Gi = []
            GiPrime = None

            if self.G2_parameters is not None:
                G2i = calculate_G2(Dij, Z, self.Rc, self.G2_parameters)
                Gi = np.append(Gi, G2i)

                if self.derivative:
                    G2iP, rG2iP = calculate_G2Prime(Rij, Dij, Z, self.total_atoms,
                                                    i, Ri, IDs, self.Rc, self.G2_parameters)
                    if GiPrime is None:
                        GiPrime = G2iP
                        if self.stress:
                            rGiPrime = rG2iP
                    else:
                        GiPrime = np.append(GiPrime, G2iP, axis=1)
                        if self.stress:
                            rGiPrime = np.append(rGiPrime, rG2iP, axis=1)

            if self.G4_parameters is not None:
                G4i = calculate_G4(Rij, IDs, Z, jks, self.Rc, self.G4_parameters)
                Gi = np.append(Gi, G4i)

                if self.derivative:
                    G4iP, rG4iP = calculate_G4Prime(Rij, Ri, i, IDs, self.total_atoms,
                                                    jks, Z, self.Rc, self.G4_parameters)

                    if GiPrime is None:
                        GiPrime = G4iP
                        if self.stress:
                            rGiPrime = rG4iP
                    else:
                        GiPrime = np.append(GiPrime, G4iP, axis=1)
                        if self.stress:
                            rGiPrime = np.append(rGiPrime, rG4iP, axis=1)

            if self.G5_parameters is not None:
                G5i = calculate_G5(Rij, IDs, Z, jks, self.Rc, self.G4_parameters)
                Gi = np.append(Gi, G5i)

                if self.derivative:
                    G5iP, rG5iP = calculate_G5Prime(Rij, Ri, i, IDs, self.total_atoms,
                                                    jks, Z, self.Rc, self.G4_parameters)
                    if GiPrime is None:
                        GiPrime = G5iP
                        if self.stress:
                            rGiPrime = rG5iP
                    else:
                        GiPrime = np.append(GiPrime, G5iP, axis=1)
                        if self.stress:
                            rGiPrime = np.append(rGiPrime, rG5iP, axis=1)

            self.all_G['x'].append(Gi)
            self.all_G['elements'].append(element)
            if self.derivative:
                self.all_G['dxdr'].append(GiPrime)
                if self.stress:
                    self.all_G['rdxdr'].append(rGiPrime)

        self.all_G['x'] = np.asarray(self.all_G['x'])
        if self.derivative:
            self.all_G['dxdr'] = np.asarray(self.all_G['dxdr'])
            if self.stress:
                self.all_G['rdxdr'] = -np.asarray(self.all_G['rdxdr']) / vol
            else:
                self.all_G['rdxdr'] = None
        return self.all_G


########################### Symmetry Functions ################################


def calculate_G2(Dij, Z, Rc, parameters):
    """Calculate radial weighted symmetry function for a center atom i.
    
    Parameters
    ----------
    Dij: float array [j]
        Array of neighbors (j) distances with respect to center atom i.
    Z: float array [j]
        Array of atomic numbers of neighbors.
    Rc: float
        The cutoff radius.
    parameters: dict
        Rs: float array (d1)
            The shift from the center of the G2 symmetry function.
        etas: float array (d2)
             eta parameters of G2 symmetry function.
    
    Returns
    -------
    G2: float array [d = d1*d2]
        G2 symmetry value.
    """
    Rs = parameters['Rs']  # [d1]
    etas = parameters['eta']  # [d2]
    d1, d2, j = len(Rs), len(etas), len(Dij)

    Z_cutoff = Cosine(Dij, Rc) * Z  # [j]
    d20 = (Dij - Rs[:, np.newaxis]) ** 2  # [d1, j]
    term = np.exp(np.einsum('i,jk->ijk', -etas, d20))  # [d2] * [d1, j] -> [d2, d1, j]
    results = np.einsum('ijk,k->ij', term, Z_cutoff)  # [d2, d1, j] * [j] -> [d2, d1]
    G2 = results.reshape([d2 * d1])  # [d2*d1]

    return G2


def calculate_G2Prime(Rij, Dij, Z, m, i, Ri, IDs, Rc, parameters):
    """Calculate the derivative of G2 symmetry function for atom i.
    
    Parameters
    ----------
    Rij: float array [j, 3]
        The vector distances of atom i to neighbors js.
    Dij: float array [j]
        Array of neighbors (j) distances with respect to center atom i.
    Z: float array [j]
        Array of atomic numbers of neighbors.
    m: int
        The total atoms in the crystal unit cell.
    i: int
        The i-th atom center.
    IDs: int array [j]
        The indices of neighbors centering about atom i.
    Ri: float array [3]
        The x, y, and z positions of atom i.
    Rc: float
        The cutoff radius.
    parameters: dict
        Rs: float array (d1)
            The shift from the center of the G2 symmetry function.
        etas: float array (d2)
             eta parameters of G2 symmetry function.
    
    Returns
    -------
    G2Prime: float array [m, d2*d1, 3]
        The derivative of G2 symmetry value at i-th atom.
    """
    Rs = parameters['Rs']  # [d1]
    etas = parameters['eta']  # [d2]
    d1, d2, j = len(Rs), len(etas), len(Rij)

    ij_list = i * np.ones([len(IDs), 2], dtype=int)
    ij_list[:, 1] = IDs  # [j]

    d0 = Dij - Rs[:, np.newaxis]  # [d1, j]
    d20 = d0 ** 2  # [d1, j]
    term1 = np.exp(np.einsum('i,jk->ijk', -etas, d20))  # [d2, d1, j]

    term21 = CosinePrime(Dij, Rc) * Z  # [j]
    Z_cutoff = Cosine(Dij, Rc) * Z  # [j]
    _term22 = np.einsum('ij,j->ij', d0, Z_cutoff)  # [d1, j]
    term22 = 2 * np.einsum('i,jk->ijk', etas, _term22)  # [d2, d1, j]
    term2 = term21 - term22  # [d2, d1, j]

    term_1_and_2 = (term1 * term2).reshape([d2 * d1, j])  # [d2*d1, j]

    dRij_dRm = np.zeros([j, 3, m])
    for mm in range(m):
        mm_list = mm * np.ones([j, 1], dtype=int)
        dRij_dRm[:, :, mm] = dRij_dRm_norm(Rij, np.hstack((ij_list, mm_list)))

    G2ip0 = np.einsum('ij,jkl->ijkl', term_1_and_2, dRij_dRm)  # [d2*d1, j, 3, m]

    rG2ip0 = np.zeros([d1 * d2, j, 3, m, 3])
    for mm, ij in enumerate(ij_list):
        _j = ij[1]
        tmp = G2ip0[:, mm, :, :]
        rG2ip0[:, mm, :, i, :] += np.einsum('ij,k->ijk', tmp[:, :, i], Ri)
        rG2ip0[:, mm, :, _j, :] += np.einsum('ij,k->ijk', tmp[:, :, _j], Rij[mm] + Ri)

    G2Prime = np.einsum('ijkl->lik', G2ip0)  # [m, d2*d1, 3]
    rG2Prime = np.einsum('ijklm->likm', rG2ip0)  # [m, d2*d1, 3, 3]

    return G2Prime, rG2Prime


def calculate_G4(Rij, IDs, Z, jks, Rc, parameters):
    """Calculate G4 symmetry function for a given atom i.
    
    Parameters
    ----------
    Rij: array [j, 3]
        The vector distances of atom i to neighbors js.
    IDs: int array [j]
        The indices of neighbors centering about atom i.
    Z: float array [j]
        Array of atomic numbers of neighbors.
    jks: int array [j*(j-1)/2, 2]
        The list of jk pairs.
    Rc: float
        The cutoff radius.
    parameters: dict
        Rs: float array (d1)
        etas: float array (d2)
        lambdas: float array (d3)
        zetas: float array (d4)
        
    Returns
    -------
    G4: array [d = d1*d2*d3*d4]
        G4 symmetry value with d members.
    """
    Rs, etas = parameters['Rs'], parameters['eta']  # [d1], [d2]
    zetas, lambdas = parameters['zeta'], parameters['lambda']  # [d3], [d4]
    d1, d2, d3, d4 = len(Rs), len(etas), len(lambdas), len(zetas)
    jk = len(jks)  # [jk]; j*k pairs

    Zjk = Z[jks[:, 0]] * Z[jks[:, 1]]  # [jk]

    rij = Rij[jks[:, 0]]  # [jk, 3]
    rik = Rij[jks[:, 1]]  # [jk, 3]
    rjk = rik - rij  # [jk, 3]
    R2ij0 = np.sum(rij ** 2., axis=1)
    R2ik0 = np.sum(rik ** 2., axis=1)
    R2jk0 = np.sum(rjk ** 2., axis=1)
    R1ij0 = np.sqrt(R2ij0)  # [jk]
    R1ik0 = np.sqrt(R2ik0)  # [jk]
    R1jk0 = np.sqrt(R2jk0)  # [jk]
    R2ij = R2ij0 - Rs[:, np.newaxis] ** 2  # [d1, jk]
    R2ik = R2ik0 - Rs[:, np.newaxis] ** 2  # [d1, jk]
    R2jk = R2jk0 - Rs[:, np.newaxis] ** 2  # [d1, jk]
    R1ij = R1ij0 - Rs[:, np.newaxis]  # [d1, jk]
    R1ik = R1ik0 - Rs[:, np.newaxis]  # [d1, jk]
    R1jk = R1jk0 - Rs[:, np.newaxis]  # [d1, jk]

    powers = 2. ** (1. - zetas)  # d4
    cos_ijk = np.sum(rij * rik, axis=1) / R1ij0 / R1ik0  # jk
    term1 = 1. + np.einsum('i,j->ij', lambdas, cos_ijk)  # [d3, jk]

    zetas1 = zetas.repeat(d3 * jk).reshape([d4, d3, jk])  # [d4, d3, jk]
    term2 = np.power(term1, zetas1)  # [d4, d3, jk]
    term3 = np.exp(np.einsum('i,jk->ijk', -etas, (R2ij + R2jk + R2ik)))  # [d2, d1, jk]
    term4 = Cosine(R1ij0, Rc) * Cosine(R1ik0, Rc) * Cosine(R1jk0, Rc) * Zjk  # [jk]
    term5 = np.einsum('ijk,lmk->ijlmk', term2, term3)  # [d4, d3, d2, d1, jk]
    term6 = np.einsum('ijkml,l->ijkml', term5, term4)  # [d4, d3, d2, d1, jk]
    results = np.einsum('i,ijkml->ijkml', powers, term6)  # [d4, d3, d2, d1, jk]
    results = results.reshape([d1 * d2 * d3 * d4, jk])  # [d4*d3*d2*d1, jk]

    G4 = np.einsum('ij->i', results)  # [d]

    return G4


def calculate_G5(Rij, IDs, Z, jks, Rc, parameters):
    """Calculate G5 symmetry function for a given atom i.
    
    Parameters
    ----------
    Rij: array [j, 3]
        The vector distances of atom i to neighbors js.
    IDs: int array [j]
        The indices of neighbors centering about atom i.
    Z: float array [j]
        Array of atomic numbers of neighbors.
    jks: int array [j*(j-1)/2, 2]
        The list of jk pairs.
    Rc: float
        The cutoff radius.
    parameters: dict
        Rs: float array (d1)
        etas: float array (d2)
        lambdas: float array (d3)
        zetas: float array (d4)
        
    Returns
    -------
    G5: array [d = d1*d2*d3*d4]
        G5 symmetry value with d members.
    """
    Rs, etas = parameters['Rs'], parameters['eta']  # [d1], [d2]
    zetas, lambdas = parameters['zeta'], parameters['lambda']  # [d3], [d4]
    d1, d2, d3, d4 = len(Rs), len(etas), len(lambdas), len(zetas)
    jk = len(jks)  # [jk]; j*k pairs

    Zjk = Z[jks[:, 0]] * Z[jks[:, 1]]  # [jk]

    rij = Rij[jks[:, 0]]  # [jk, 3]
    rik = Rij[jks[:, 1]]  # [jk, 3]
    R2ij0 = np.sum(rij ** 2., axis=1)
    R2ik0 = np.sum(rik ** 2., axis=1)

    R1ij0 = np.sqrt(R2ij0)  # [jk]
    R1ik0 = np.sqrt(R2ik0)  # [jk]

    R2ij = R2ij0 - Rs[:, np.newaxis] ** 2  # [d1, jk]
    R2ik = R2ik0 - Rs[:, np.newaxis] ** 2  # [d1, jk]

    R1ij = R1ij0 - Rs[:, np.newaxis]  # [d1, jk]
    R1ik = R1ik0 - Rs[:, np.newaxis]  # [d1, jk]

    powers = 2. ** (1. - zetas)  # [d4]
    cos_ijk = np.sum(rij * rik, axis=1) / R1ij0 / R1ik0  # [jk]
    term1 = 1. + np.einsum('i,j->ij', lambdas, cos_ijk)  # [d3, jk]

    zetas1 = zetas.repeat(d3 * jk).reshape([d4, d3, jk])  # [d4, d3, jk]
    term2 = np.power(term1, zetas1)  # [d4, d3, jk]
    term3 = np.exp(np.einsum('i,jk->ijk', -etas, (R2ij + R2ik)))  # [d2, d1, jk]
    term4 = Cosine(R1ij0, Rc) * Cosine(R1ik0, Rc) * Zjk  # [jk]
    term5 = np.einsum('ijk,lmk->ijlmk', term2, term3)  # [d4, d3, d2, d1, jk]
    term6 = np.einsum('ijkml,l->ijkml', term5, term4)  # [d4, d3, d2, d1, jk]
    results = np.einsum('i,ijkml->ijkml', powers, term6)  # [d4, d3, d2, d1, jk]
    results = results.reshape([d1 * d2 * d3 * d4, jk])  # [d, jk]

    G5 = np.einsum('ij->i', results)  # [d]

    return G5


def calculate_G4Prime(Rij, Ri, i, IDs, m, jks, Z, Rc, parameters):
    """Calculate G4 symmetry function for a given atom i.
    
    Parameters
    ----------
    Rij: array [j, 3]
        The vector distances of atom i to neighbors js.
    Ri: float array [3]
        The x, y, and z positions of atom i.
    i: int
        The i-th atom center.    
    IDs: int array [j]
        The indices of neighbors centering about atom i.
    m: int
        The total atoms in the crystal unit cell.
    jks: int array [j*k, 2]
        The list of [j,k] pairs.
    Z: float array [j]
        Array of atomic numbers of neighbors.
    Rc: float
        The cutoff radius.
    parameters: dict
        Rs: float array (d1)
        etas: float array (d2)
        lambdas: float array (d3)
        zetas: float array (d4)
        
    Returns
    -------
        G4Prime: array [m, d, 3]
        The derivative of G4 symmetry value at i-th atom. m is the index of the
        atom that force is acting on.
    """
    Rs, etas = parameters['Rs'], parameters['eta']  # [d1], [d2]
    zetas, lambdas = parameters['zeta'], parameters['lambda']  # [d3], [d4]
    d1, d2, d3, d4 = len(Rs), len(etas), len(lambdas), len(zetas)
    jk = len(jks)  # [jk]; j*k pairs

    ijk_list = i * np.ones([jk, 3], dtype=int)
    ijk_list[:, 1] = IDs[jks[:, 0]]
    ijk_list[:, 2] = IDs[jks[:, 1]]

    Zjk = Z[jks[:, 0]] * Z[jks[:, 1]]  # [jk]

    rij = Rij[jks[:, 0]]  # [jk, 3]
    rik = Rij[jks[:, 1]]  # [jk, 3]
    rjk = rik - rij  # [jk, 3]

    R2ij0 = np.sum(rij ** 2., axis=1)
    R2ik0 = np.sum(rik ** 2., axis=1)
    R2jk0 = np.sum(rjk ** 2., axis=1)
    R1ij0 = np.sqrt(R2ij0)  # jk
    R1ik0 = np.sqrt(R2ik0)  # jk
    R1jk0 = np.sqrt(R2jk0)  # jk
    R2ij = R2ij0 - Rs[:, np.newaxis] ** 2  # [d1, jk]
    R2ik = R2ik0 - Rs[:, np.newaxis] ** 2  # [d1, jk]
    R2jk = R2jk0 - Rs[:, np.newaxis] ** 2  # [d1, jk]
    R1ij = R1ij0 - Rs[:, np.newaxis]  # [d1, jk]
    R1ik = R1ik0 - Rs[:, np.newaxis]  # [d1, jk]
    R1jk = R1jk0 - Rs[:, np.newaxis]  # [d1, jk]

    cos_ijk = np.sum(rij * rik, axis=1) / R1ij0 / R1ik0  # [jk]

    dfcij = CosinePrime(R1ij0, Rc)
    dfcjk = CosinePrime(R1jk0, Rc)
    dfcik = CosinePrime(R1ik0, Rc)
    fcij = Cosine(R1ij0, Rc)
    fcjk = Cosine(R1jk0, Rc)
    fcik = Cosine(R1ik0, Rc)

    powers = 2. ** (1. - zetas)  # d4
    term1 = 1. + np.einsum('i,j->ij', lambdas, cos_ijk)  # [d3, jk]
    zetas1 = zetas.repeat(d3 * jk).reshape([d4, d3, jk])  # [d4, d3, jk]
    term2 = np.power(term1, zetas1 - 1)  # [d4, d3, jk]
    g41 = np.exp(np.einsum('i,jk->ijk', -etas, (R2ij + R2jk + R2ik)))  # [d2, d1, jk]
    g41 = np.einsum('ijk,lmk->ijlmk', term2, g41)  # [d4, d3, d2, d1, jk]
    g41 = np.einsum('i,ijklm->ijklm', powers, g41)  # [d4, d3, d2, d1, jk]

    lambda_zeta = np.einsum('i,j->ij', zetas, lambdas)  # [d4, d3]
    (dRij_dRm, dRik_dRm, dRjk_dRm) = dRijk_dRm(rij, rik, rjk, ijk_list, m)  # [jk, 3, m]
    Rijk_dRm = np.einsum('i,ijk->ijk', R1ij0, dRij_dRm) + \
               np.einsum('i,ijk->ijk', R1ik0, dRik_dRm) + \
               np.einsum('i,ijk->ijk', R1jk0, dRjk_dRm)
    dcos = dcosijk_dRm(rij, rik, ijk_list, dRij_dRm, dRik_dRm)
    dcos = np.einsum('ij,klm->ijklm', lambda_zeta, dcos)  # [d4, d3, 3, jk, m]
    dcos = np.broadcast_to(dcos, (d2,) + (d4, d3, jk, 3, m))
    dcos = np.transpose(dcos, (1, 2, 0, 3, 4, 5))
    cost = np.einsum('i,jk->jik', 2 * etas, term1)  # [d3, d2, jk]
    cost = np.einsum('ijk,klm->ijklm', cost, Rijk_dRm)  # [d3, d2, jk, 3, m]
    cost = np.broadcast_to(cost, (d4,) + (d3, d2, jk, 3, m))
    g42 = np.einsum('l,ijklmn->ijklmn', fcij * fcik * fcjk * Zjk, dcos - cost)  # [d4, d3, d2, jk, 3, m]

    g43 = np.einsum('i,ijk->ijk', dfcij * fcik * fcjk * Zjk, dRij_dRm) + \
          np.einsum('i,ijk->ijk', fcij * dfcik * fcjk * Zjk, dRik_dRm) + \
          np.einsum('i,ijk->ijk', fcij * fcik * dfcjk * Zjk, dRjk_dRm)
    g43 = np.einsum('ij,jkl->ijkl', term1, g43)  # [d3, jk, 3, m]
    g43 = np.broadcast_to(g43, (d4, d2,) + (d3, jk, 3, m))
    g43 = np.transpose(g43, (0, 2, 1, 3, 4, 5))

    # [d4, d3, d2, d1, jk] * [d4, d3, d2, jk, 3, m] -> [d4, d3, d2, d1, jk, 3, m] -> [S, jk, 3, m] 
    G4ip0 = np.einsum('ijklm, ijkmno->ijklmno', g41, g42 + g43, \
                      optimize='greedy').reshape([d1 * d2 * d3 * d4, jk, 3, m])
    # [S, m, 3, N] * [m, 3] -> [S, m, 3, N, 3] 
    # partition the dxdr to each i, j, k
    rG4ip0 = np.zeros([d1 * d2 * d3 * d4, len(ijk_list), 3, m, 3])
    for mm, ijk in enumerate(ijk_list):
        j, k = ijk[1], ijk[2]
        tmp = G4ip0[:, mm, :, :]  # S,m,3 -> S,3 * 3 -> S*3*3
        rG4ip0[:, mm, :, i, :] += np.einsum('ij,k->ijk', tmp[:, :, i], Ri)
        rG4ip0[:, mm, :, j, :] += np.einsum('ij,k->ijk', tmp[:, :, j], rij[mm] + Ri)
        rG4ip0[:, mm, :, k, :] += np.einsum('ij,k->ijk', tmp[:, :, k], rik[mm] + Ri)

    G4Prime = np.einsum('ijkl->lik', G4ip0)
    rG4Prime = np.einsum('ijklm->likm', rG4ip0)

    return G4Prime, rG4Prime


def calculate_G5Prime(Rij, Ri, i, IDs, m, jks, Z, Rc, parameters):
    """Calculate G5 symmetry function for a given atom i.
    
    Parameters
    ----------
    Rij: array [j, 3]
        The vector distances of atom i to neighbors js.
    Ri: float array [3]
        The x, y, and z positions of atom i.
    i: int
        The i-th atom center.    
    IDs: int array [j]
        The indices of neighbors centering about atom i.
    m: int
        The total atoms in the crystal unit cell.
    jks: int array [j*k, 2]
        The list of [j,k] pairs.
    Z: float array [j]
        Array of atomic numbers of neighbors.
    Rc: float
        The cutoff radius.
    parameters: dict
        Rs: float array (d1)
        etas: float array (d2)
        lambdas: float array (d3)
        zetas: float array (d4)
        
    Returns
    -------
    G5Prime: array [m, d, 3]
        The derivative of G5 symmetry value at i-th atom. m is the index of the
        atom that force is acting on.
    """
    Rs, etas = parameters['Rs'], parameters['eta']  # [d1], [d2]
    zetas, lambdas = parameters['zeta'], parameters['lambda']  # [d3], [d4]
    d1, d2, d3, d4 = len(Rs), len(etas), len(lambdas), len(zetas)
    jk = len(jks)  # [jk]; j*k pairs

    ijk_list = i * np.ones([jk, 3], dtype=int)
    ijk_list[:, 1] = IDs[jks[:, 0]]
    ijk_list[:, 2] = IDs[jks[:, 1]]

    Zjk = Z[jks[:, 0]] * Z[jks[:, 1]]  # [jk]

    rij = Rij[jks[:, 0]]  # [jk, 3]
    rik = Rij[jks[:, 1]]  # [jk, 3]
    rjk = rik - rij  # [jk, 3]

    R2ij0 = np.sum(rij ** 2., axis=1)
    R2ik0 = np.sum(rik ** 2., axis=1)
    R1ij0 = np.sqrt(R2ij0)  # [jk]
    R1ik0 = np.sqrt(R2ik0)  # [jk]
    R2ij = R2ij0 - Rs[:, np.newaxis] ** 2  # [d1, jk]
    R2ik = R2ik0 - Rs[:, np.newaxis] ** 2  # [d1, jk]
    R1ij = R1ij0 - Rs[:, np.newaxis]  # [d1, jk]
    R1ik = R1ik0 - Rs[:, np.newaxis]  # [d1, jk]

    cos_ijk = np.sum(rij * rik, axis=1) / R1ij0 / R1ik0  # [jk]

    dfcij = CosinePrime(R1ij0, Rc)
    dfcik = CosinePrime(R1ik0, Rc)
    fcij = Cosine(R1ij0, Rc)
    fcik = Cosine(R1ik0, Rc)

    powers = 2. ** (1. - zetas)  # [d4]
    term1 = 1. + np.einsum('i,j->ij', lambdas, cos_ijk)  # [d3, jk]
    zetas1 = zetas.repeat(d3 * jk).reshape([d4, d3, jk])  # [d4, d3, jk]
    term2 = np.power(term1, zetas1 - 1)  # [d4, d3, jk]
    g51 = np.exp(np.einsum('i,jk->ijk', -etas, (R2ij + R2ik)))  # [d2, d1, jk]
    g51 = np.einsum('ijk,lmk->ijlmk', term2, g51)  # [d4, d3, d2, d1, jk]
    g51 = np.einsum('i,ijklm->ijklm', powers, g51)  # [d4, d3, d2, d1, jk]

    lambda_zeta = np.einsum('i,j->ij', zetas, lambdas)  # [d4, d3]
    (dRij_dRm, dRik_dRm, dRjk_dRm) = dRijk_dRm(rij, rik, rjk, ijk_list, m)  # [jk, 3, m]
    Rijk_dRm = np.einsum('i,ijk->ijk', R1ij0, dRij_dRm) + \
               np.einsum('i,ijk->ijk', R1ik0, dRik_dRm)
    dcos = dcosijk_dRm(rij, rik, ijk_list, dRij_dRm, dRik_dRm)
    dcos = np.einsum('ij,klm->ijklm', lambda_zeta, dcos)  # [d4, d3, 3, jk, m]
    dcos = np.broadcast_to(dcos, (d2,) + (d4, d3, jk, 3, m))
    dcos = np.transpose(dcos, (1, 2, 0, 3, 4, 5))
    cost = np.einsum('i,jk->jik', 2 * etas, term1)  # [d3, d2, jk]
    cost = np.einsum('ijk,klm->ijklm', cost, Rijk_dRm)  # [d3, d2, jk, 3, m]
    cost = np.broadcast_to(cost, (d4,) + (d3, d2, jk, 3, m))
    g52 = np.einsum('l,ijklmn->ijklmn', fcij * fcik * Zjk, dcos - cost)  # [d4, d3, d2, jk, 3, m]

    g53 = np.einsum('i,ijk->ijk', dfcij * fcik * Zjk, dRij_dRm) + \
          np.einsum('i,ijk->ijk', fcij * dfcik * Zjk, dRik_dRm)
    g53 = np.einsum('ij,jkl->ijkl', term1, g53)  # [d3, jk, 3,m]
    g53 = np.broadcast_to(g53, (d4, d2,) + (d3, jk, 3, m))
    g53 = np.transpose(g53, (0, 2, 1, 3, 4, 5))

    G5ip0 = np.einsum('ijklm, ijkmno->ijklmno', g51, g52 + g53, \
                      optimize='greedy').reshape([d1 * d2 * d3 * d4, jk, 3, m])

    # [S, m, 3, N] * [m, 3] -> [S, m, 3, N, 3] 
    # partition the dxdr to each i, j, k
    rG5ip0 = np.zeros([d1 * d2 * d3 * d4, len(ijk_list), 3, m, 3])
    for mm, ijk in enumerate(ijk_list):
        j, k = ijk[1], ijk[2]
        tmp = G5ip0[:, mm, :, :]  # S,N,3 -> S,3 * 3 -> S*3*3
        rG5ip0[:, mm, :, i, :] += np.einsum('ij,k->ijk', tmp[:, :, i], Ri)
        rG5ip0[:, mm, :, j, :] += np.einsum('ij,k->ijk', tmp[:, :, j], rij[mm] + Ri)
        rG5ip0[:, mm, :, k, :] += np.einsum('ij,k->ijk', tmp[:, :, k], rik[mm] + Ri)

    G5Prime = np.einsum('ijkl->lik', G5ip0)
    rG5Prime = np.einsum('ijklm->likm', rG5ip0)

    return G5Prime, rG5Prime


def dRij_dRm_norm(Rij, ijm_list):
    """Calculate the derivative of Rij norm w. r. t. atom m. This term affects 
    only on i and j.
    
    Parameters
    ----------
    Rij : array [j, 3] or [j*k, 3]
        The vector distances of atom i to atom j.
    ijm_list: array [j, 3] or [j*k, 3]
        Id list of center atom i, neighbors atom j, and atom m.
    
    Returns
    -------
    dRij_m: array [j, 3] or [j*k, 3]
        The derivative of pair atoms w.r.t. atom m in x, y, z directions.
    """
    dRij_m = np.zeros([len(Rij), 3])
    R1ij = np.linalg.norm(Rij, axis=1).reshape([len(Rij), 1])
    l1 = (ijm_list[:, 2] == ijm_list[:, 0])
    dRij_m[l1, :] = -Rij[l1] / R1ij[l1]
    l2 = (ijm_list[:, 2] == ijm_list[:, 1])
    dRij_m[l2, :] = Rij[l2] / R1ij[l2]
    l3 = (ijm_list[:, 0] == ijm_list[:, 1])
    dRij_m[l3, :] = 0

    return dRij_m


def dcosijk_dRm(Rij, Rik, ijk_list, dRij_dRm, dRik_dRm):
    """Calculate the derivative of cosine_ijk function w. r. t. atom m.
    m must belong to one of the ijks. Otherwise, the derivative is zero.
    If the input Rij and Rik are (j*k)*3 dimensions, the output will be 
    (j*k)*3*3. The extra dimension comes from looping over {i, j, k}.
    
    Parameters
    ----------
    Rij: array [j*k, 3]
        The vector distances of atom i to neighbors js.
    Rik: array [j*k, 3]
        The vector distances of atom i to neighbors ks, where j != k.
        
    Returns
    -------
    Derivative of cosine dot product w.r.t. the radius of an atom m,
    The atom m has to be in the an array with 3 indices: i, j, and k.
    """
    m = dRik_dRm.shape[-1]
    Dij = np.linalg.norm(Rij, axis=1)
    Dik = np.linalg.norm(Rik, axis=1)
    rDijDik = 1 / Dij / Dik  # jk array
    Rij_Rik = np.sum(Rij * Rik, axis=1)  # jk array
    ij_list = ijk_list[:, [0, 1]]
    ik_list = ijk_list[:, [0, 2]]
    dcos = np.zeros([len(Rij), 3, m])

    for mm in range(m):
        mm_list = mm * np.ones([len(Rij), 1], dtype=int)
        dRij_dRm_v = dRij_dRm_vector(np.append(ij_list, mm_list, axis=1))
        dRik_dRm_v = dRij_dRm_vector(np.append(ik_list, mm_list, axis=1))
        term10 = np.einsum('i,ij->ij', rDijDik, Rik)  # jk*3 arrray
        dcos[:, :, mm] += np.einsum('ij,i->ij', term10, dRij_dRm_v)
        term20 = np.einsum('i,ij->ij', rDijDik, Rij)  # jk*3 array
        dcos[:, :, mm] += np.einsum('ij,i->ij', term20, dRik_dRm_v)
        term30 = Rij_Rik * rDijDik / Dij  # jk*3
        dcos[:, :, mm] -= np.einsum('i,ij->ij', term30, dRij_dRm[:, :, mm])
        term40 = Rij_Rik * rDijDik / Dik  # jk*3
        dcos[:, :, mm] -= np.einsum('i,ij->ij', term40, dRik_dRm[:, :, mm])  # jk*3*m

    return dcos


def dRij_dRm_vector(ijm_list):
    """Calculate the derivative of Rij vector w. r. t. atom m.
    
    Parameters
    ----------
    ijm_list: array [i*j, 3]
        List of indices of center atom i, neighbors atom j, and atom m.
    
    Returns
    -------
    list of float
        The derivative of the position vector R_{ij} with respect to atom
        index m in x, y, z directions.
    """

    dRij_dRm_vector = np.zeros(len(ijm_list))
    l1 = (ijm_list[:, 2] == ijm_list[:, 0])
    dRij_dRm_vector[l1] = -1
    l2 = (ijm_list[:, 2] == ijm_list[:, 1])
    dRij_dRm_vector[l2] = 1

    return dRij_dRm_vector


def dRijk_dRm(Rij, Rik, Rjk, ijk_list, m):
    """Calculate the derivative of R_{ab} norm with respect to atom m.
    
    Parameters
    ----------
    Rij: array [j*k, 3]
        The vector distances of atom i to neighbors js.
    Rik: array [j*k, 3]
        The vector distances of atom i to neighbors ks.
    Rjk: array [j*k, 3]
        The vector distances of atom j to atom k, where j != k.
    ijk_list: array
        The combinations of atom i, neighbor j, and neighbor k indices.
    m: int
        the total number of atom in the unit cell
        
    Returns
    -------
    dR{ab}_dRm: array [j*k, 3, m]
        The derivative of R_{ab} norm with respect to atom m.
    """
    dRij_dRm = np.zeros([len(Rij), 3, m])
    dRik_dRm = np.zeros([len(Rij), 3, m])
    dRjk_dRm = np.zeros([len(Rij), 3, m])
    ij_list = ijk_list[:, [0, 1]]
    ik_list = ijk_list[:, [0, 2]]
    jk_list = ijk_list[:, [1, 2]]

    for mm in range(m):
        mm_list = mm * np.ones([len(Rij), 1], dtype=int)
        dRij_dRm[:, :, mm] = dRij_dRm_norm(Rij, np.append(ij_list, mm_list,
                                                          axis=1))
        dRik_dRm[:, :, mm] = dRij_dRm_norm(Rik, np.append(ik_list, mm_list,
                                                          axis=1))
        dRjk_dRm[:, :, mm] = dRij_dRm_norm(Rjk, np.append(jk_list, mm_list,
                                                          axis=1))

    return (dRij_dRm, dRik_dRm, dRjk_dRm)


############################# Cutoff Functionals ##############################


def Cosine(Rij, Rc):
    # Rij is the norm 
    ids = (Rij > Rc)
    result = 0.5 * (np.cos(np.pi * Rij / Rc) + 1.)
    result[ids] = 0

    return result


def CosinePrime(Rij, Rc):
    # Rij is the norm
    ids = (Rij > Rc)
    result = -0.5 * np.pi / Rc * np.sin(np.pi * Rij / Rc)
    result[ids] = 0

    return result


########################### Auxiliary Functions ###############################


def to_array(input):
    if isinstance(input, (list, np.ndarray)):
        res = np.asarray(input)
    else:
        res = np.asarray([input])
    return res


def create_type_set(number_set, order=1):
    types = list(set(number_set))
    return np.array(list(combinations_with_replacement(types, order)))


def select_rows(data, row_pattern):
    if len(row_pattern) == 1:
        ids = (data == row_pattern)
    elif len(row_pattern) == 2:
        a, b = row_pattern
        ids = []
        for id, d in enumerate(data):
            if a == b:
                if d[0] == a and d[1] == a:
                    ids.append(id)
            else:
                if (d[0] == a and d[1] == b) or (d[0] == b and d[1] == a):
                    ids.append(id)
    return ids


################################### Test ######################################


if __name__ == '__main__':
    from ase.build import bulk

    np.set_printoptions(formatter={'float': '{: 0.4f}'.format})

    # Set up symmetry parameters
    Rc = 5.5
    symmetry = {'G2': {'eta': [0.036, 0.071, ], 'Rs': [0]},
                'G4': {'Rs': [0], 'lambda': [1], 'zeta': [1, ], 'eta': [0.036, 0.071]},
                'G5': {'Rs': [0], 'lambda': [1], 'zeta': [1, ], 'eta': [0.036, 0.071]},
                }

    for a in [5.0]:
        si = bulk('Si', 'diamond', a=a, cubic=True)
        cell = si.get_cell()
        si.set_cell(cell)
        print(si.get_cell())

        bp = wACSF(symmetry, Rc=Rc, derivative=True, stress=True)
        des = bp.calculate(si)
        print("G:", des['x'][0])
        print("GPrime", des['dxdr'][0, :, 0, :])
        print("GPrime", des['dxdr'][0, :, 2, :])
        print("GPrime", des['dxdr'][0, :, 4, :])
        print(np.einsum('ijklm->klm', des['rdxdr']))



from __future__ import division

from optparse import OptionParser

import numba as nb
import numpy as np
from ase.neighborlist import NeighborList
from numba import prange

from .angular_momentum import Wigner_D, factorial, deltacg, Wigner_D_wDerivative


class SO4_Bispectrum:
    '''
    Pyxtal implementation of SO4 bispectrum component calculator.
    The difference between this implementation and the SNAP implementation
    lies exclusively with the choice of unit quaternion (removing singularities
    for rotations of 0 and 2pi) and the method of calculating Wigner-U functions

    here we use a polynomial form of the Wigner-D matrices to calculate the U-functions
    and thus the gradients can be calculated simultaneously through differentiating the
    U-functions using horner form
    '''

    def __init__(self, lmax=3, rcut=3, derivative=True, stress=False, normalize_U=False):
        # populate attributes
        self.lmax = lmax
        self.rcut = rcut
        self.derivative = derivative
        self.stress = stress
        self.normalize_U = normalize_U

    def __repr__(self):
        return """S04 bispectrum coefficient calculator for the
    atomic neighbor density function for a Wigner-D
    expansion of order {} and cutoff
    radius {}""".format(self.lmax, self.rcut)

    @property
    def lmax(self):
        return self._twol // 2

    @lmax.setter
    def lmax(self, lmax):
        if isinstance(lmax, int) is True:
            if lmax < 0:
                raise ValueError('lmax must be greater than or equal to zero')
            elif lmax > 32:
                raise NotImplementedError('''Currently we only support Wigner-D matrices and spherical harmonics
                                          for arguments up to l=32.  If you need higher functionality, raise an issue
                                          in our github and we will expand the set of supported functions''')
            self._twol = 2 * lmax
        else:
            raise ValueError('lmax must be an integer')

    @property
    def rcut(self):
        return self._rcut

    @rcut.setter
    def rcut(self, rcut):
        if isinstance(rcut, float) is True or isinstance(rcut, int) is True:
            if rcut <= 0:
                raise ValueError('rcut must be greater than zero')
            self._rcut = rcut
        else:
            raise ValueError('rcut must be a float')

    @property
    def derivative(self):
        return self._derivative

    @derivative.setter
    def derivative(self, derivative):
        if isinstance(derivative, bool) is True:
            self._derivative = derivative
        else:
            raise ValueError('derivative must be a boolean value')

    @property
    def stress(self):
        return self._stress

    @stress.setter
    def stress(self, stress):
        if isinstance(stress, bool) is True:
            self._stress = stress
        else:
            raise ValueError('stress must be a boolean value')

    @property
    def normalize_U(self):
        return self._norm

    @normalize_U.setter
    def normalize_U(self, normalize_U):
        if isinstance(normalize_U, bool) is True:
            self._norm = normalize_U
        else:
            raise ValueError('normalize_U must be a boolean value')

    def clear_memory(self):
        '''
        Clears all memory that isn't an essential attribute for the calculator
        '''
        attrs = list(vars(self).keys())
        for attr in attrs:
            if attr not in {'_twol', '_rcut', '_derivative', '_stress', '_norm'}:
                delattr(self, attr)
        return

    def calculate(self, atoms):
        '''
        args:
            atoms:  ASE atoms object for the corresponding structure

        returns: a dictionary with the bispectrum components, their
        gradients, and the elemental specie of each atom in the atoms
        object
        '''
        self._atoms = atoms
        vol = atoms.get_volume()
        self.build_neighbor_list()
        self.initialize_arrays()

        get_bispectrum_components(self.center_atoms, self.neighborlist, self.seq,
                                  self.atomic_numbers, self.site_atomic_numbers,
                                  self._twol, self.rcut, self._norm, self.derivative,
                                  self.stress, self._blist, self._dblist, self._bstress)

        if self.derivative is True:

            x = {'x': self._blist.real, 'dxdr': self._dblist.real,
                 'elements': list(atoms.symbols), 'seq': self.seq}

            if self.stress is True:
                x['rdxdr'] = self._bstress.real / vol
            else:
                x['rdxdr'] = None

        else:
            x = {'x': self._blist.real, 'dxdr': None, 'elements': list(atoms.symbols)}

            if self.stress is True:
                x['rdxdr'] = self._bstress.real / vol
            else:
                x['rdxdr'] = None

        self.clear_memory()
        return x

    def build_neighbor_list(self):
        '''
        Builds a neighborlist for the calculation of bispectrum components for
        a given ASE atoms object given in the calculate method.
        '''
        atoms = self._atoms
        cutoffs = [self.rcut / 2] * len(atoms)
        nl = NeighborList(cutoffs, self_interaction=False, bothways=True, skin=0.0)
        nl.update(atoms)
        center_atoms = np.zeros((len(atoms), 3), dtype=np.float64)
        neighbors = []
        neighbor_indices = []
        atomic_numbers = []

        for i in range(len(atoms)):
            # get center atom position
            center_atom = atoms.positions[i]
            center_atoms[i] = center_atom
            # get indices and cell offsets of each neighbor
            indices, offsets = nl.get_neighbors(i)
            # add an empty list to neighbors and atomic numbers for population
            neighbors.append([])
            atomic_numbers.append([])
            # the indices are already numpy arrays so just append as is
            neighbor_indices.append(indices)
            for j, offset in zip(indices, offsets):
                # compute separation vector
                pos = atoms.positions[j] + np.dot(offset, atoms.get_cell()) - center_atom
                neighbors[i].append(pos)
                atomic_numbers[i].append(atoms[j].number)

        Neighbors = []
        Atomic_numbers = []
        Seq = []
        max_len = 0
        for i in range(len(atoms)):
            unique_atoms = np.unique(neighbor_indices[i])
            if i not in unique_atoms:
                at = list(unique_atoms)
                at.append(i)
                at.sort()
                unique_atoms = np.array(at)
            for j in unique_atoms:
                Seq.append([i, j])
                neigh_locs = neighbor_indices[i] == j
                temp_neighs = np.array(neighbors[i])
                temp_ANs = np.array(atomic_numbers[i])
                Neighbors.append(temp_neighs[neigh_locs])
                Atomic_numbers.append(temp_ANs[neigh_locs])
                length = sum(neigh_locs)
                if length > max_len:
                    max_len = length

        neighborlist = np.zeros((len(Neighbors), max_len, 3), dtype=np.float64)
        Seq = np.array(Seq, dtype=np.int64)
        atm_nums = np.zeros((len(Neighbors), max_len), dtype=np.int64)
        site_atomic_numbers = np.array(list(atoms.numbers), dtype=np.int64)

        for i in range(len(Neighbors)):
            neighborlist[i, :len(Neighbors[i]), :] = Neighbors[i]
            atm_nums[i, :len(Neighbors[i])] = Atomic_numbers[i]

        # assign these arrays to attributes
        self.center_atoms = center_atoms
        self.neighborlist = neighborlist
        self.seq = Seq
        self.atomic_numbers = atm_nums
        self.site_atomic_numbers = site_atomic_numbers

        return

    def initialize_arrays(self):
        '''
        Initialize the arrays to store the bispectrum components and their
        derivatives
        '''
        # number of atoms in periodic arrangement
        ncell = len(self._atoms)
        # degree of hyperspherical expansion
        lmax = self.lmax
        # number of bispectrum coefficents
        ncoefs = round(int((lmax + 1) * (lmax + 2) * (lmax + 1.5) // 3))
        # allocate memory for the bispectrum and its derivative
        self._blist = np.zeros([ncell, ncoefs], dtype=np.complex128)
        self._dblist = np.zeros([len(self.seq), ncoefs, 3], dtype=np.complex128)
        self._bstress = np.zeros([len(self.seq), ncoefs, 3, 3], dtype=np.complex128)
        return


@nb.njit(nb.void(nb.i8, nb.f8[:]), cache=True,
         fastmath=True, nogil=True)
def init_clebsch_gordan(twol, cglist):
    '''Populate an array ofpe Clebsch-Gordan coefficients needed
    for a bispectrum coefficient calculation

    Clebsch-Gordan coefficients arise in the coupling of
    angular momenta.  The method to calculate here is given in
    "Quantum Theory of Angular Momentum" D.A. Varshalovich. 8.2.1 (3)

    Parameters
    ----------
    twol:  integer
        Corresponds to the order of hyperspherical expansion
        in this software, the degree of expansion is defined and handled
        within the SO4_Bispectrum class.
    cglist:  1-D array for storing the coefficients.  The sizing of
        this array is handled in the SO4_Bispectrum class.

    Returns
    -------
    None
    '''
    ldim = twol + 1

    idxcg_count = 0
    for l1 in range(0, ldim, 1):
        for l2 in range(0, l1 + 1, 1):
            for l in range(l1 - l2, min(twol, l1 + l2) + 1, 2):
                for m1 in range(0, l1 + 1, 1):
                    aa2 = 2 * m1 - l1
                    for m2 in range(0, l2 + 1, 1):
                        bb2 = 2 * m2 - l2
                        m = (aa2 + bb2 + l) / 2

                        if (m < 0 or m > l):
                            cglist[idxcg_count] = 0.0
                            idxcg_count += 1
                            continue

                        Sum = 0.0

                        for z in range(max(0, max(-(l - l2 + aa2) // 2, -(l - l1 - bb2) // 2)), min(
                                (l1 + l2 - l) // 2, min((l1 - aa2) // 2, (l2 + bb2) // 2)) + 1, 1):

                            if z % 2 == 0:
                                ifac = 1
                            else:
                                ifac = -1

                            Sum += ifac / (factorial(z) *
                                           factorial((l1 + l2 - l) // 2 - z) *
                                           factorial((l1 - aa2) // 2 - z) *
                                           factorial((l2 + bb2) // 2 - z) *
                                           factorial((l - l2 + aa2) // 2 + z) *
                                           factorial((l - l1 - bb2) // 2 + z))

                        cc2 = 2 * m - l
                        dcg = deltacg(l1, l2, l)
                        sfaccg = np.sqrt(factorial((l1 + aa2) // 2) *
                                         factorial((l1 - aa2) // 2) *
                                         factorial((l2 + bb2) // 2) *
                                         factorial((l2 - bb2) // 2) *
                                         factorial((l + cc2) // 2) *
                                         factorial((l - cc2) // 2) *
                                         (l + 1))

                        cglist[idxcg_count] = Sum * dcg * sfaccg
                        idxcg_count += 1

    return


@nb.njit(nb.f8(nb.f8, nb.f8), cache=True, fastmath=True, nogil=True)
def compute_sfac(r, rcut):
    '''Calculates the cosine cutoff function value given in
    On Representing Chemical Environments, Batrok, et al.

    The cosine cutoff function ensures that the hyperspherical
    expansion for the calculation of bispectrum coefficients goes
    smoothly to zero for atomic neighbors tending to the cutoff
    radius.

    Parameters
    ----------
    r: float
        The magnitude of the separation vector from
        an atom in the unit cell to a particular neighbor
    rcut: float
        The cutoff radius specified in the SO4_Bispectrum class

    Returns
    -------
    cosine_cutoff:  float 0.5 * (cos(pi*r/rcut) + 1.0)
    '''
    if r > rcut:
        return 0
    else:
        rcutfac = np.pi / rcut
        return 0.5 * (np.cos(r * rcutfac) + 1.0)


@nb.njit(nb.f8(nb.f8, nb.f8), cache=True, fastmath=True, nogil=True)
def compute_dsfac(r, rcut):
    '''Calculates the derivative of the cosine cutoff for a given radii

    Parameters
    ----------
    r: float
        see compute_sfac
    rcut: float
        see compute_sfac

    Returns
    -------
    dcosine_cutoff/dr float -0.5*pi/rcut*sin(pi*r/rcut)
    '''
    if r > rcut:
        return 0

    else:
        rcutfac = np.pi / rcut
        return -0.5 * np.sin(r * rcutfac) * rcutfac


@nb.njit(nb.void(nb.i8, nb.i8[:], nb.c16[:, :]), cache=True,
         fastmath=True, nogil=True)
def addself_uarraytot(twol, idxu_block, ulisttot):
    '''Add the central atom contribution to the hyperspherical
    expansion coefficient array.

    This initializes the expansion coefficient array with a
    Kroenocker delta for ma and mb.

    Parameters
    ----------
    twol: integer
        Corresponds to the order of hyperspherical expansion
        in this software, the degree of expansion is defined
        and handled within the SO4 bispectrum class.

    idxu_block: 1-D integer array
        Each element corresponds to the first element of the
        expansion coefficient array ->(l,0,0) for each l.  This
        is handled in the SO4_Bispectrum class.

    ulisttot: 1-D complex array
        Array for storing hyperspherical expansion coefficients.
        The sizing and filling of this array is handled in the
        SO4_Bispectrum class.

    Returns
    -------
    None
    '''
    ldim = twol + 1
    for l in range(0, ldim, 1):
        llu = idxu_block[l]
        for ma in range(0, l + 1, 1):
            ulisttot[llu] = 1.0 + 0.0j
            llu += l + 2
    return


@nb.njit(nb.c16(nb.i8, nb.i8, nb.i8, nb.c16, nb.c16),
         cache=True, fastmath=True, nogil=True)
def U(l, mp, m, a, b):
    '''Computes the hyperspherical harmonic function
    value for given Cayley-Klein parameters and angular
    momentum numbers.  The hyperspherical harmonics are the
    elements of the Wigner-D matrices.

    This function is an interface to the Wigner-D matrix
    function in angular_momentum.py.  The indexing and
    generation of the Cayley-Klein parameters is handled in
    the compute_uarray functions, which are handled by
    the SO4 bispectrum class.

    Parameters
    ----------
        l:  integer
        positive integer parameter, corresponds to 2 times the
        orbital angular momentum quantum number.

        mp:  integer
        integer parameter, corresponds to the magnetic quantum
        number by the relation m' = mp - l/2

        m:  integer
        integer parameter, corresponds to the magnetic quantum
        number by the relation m = m - l/2

        a:  complex
        Cayley-Klein parameter of unit quaternion.  Corresponds
        to the diagonal elements in the corresponding element of
        SU(2) in the 2x2 matrix representation.

        b:  complex
        Cayley-Klein parameter of unit quaternion.  Corresponds
        to the off-diagonal elements in the corresponding element of
        SU(2) in the 2x2 matrix representation.


    Returns
    -------
    complex, element of the Wigner-D matrix, see
    the corresponding function in angular_momentum.py
    '''
    m = m - l / 2
    mp = mp - l / 2
    m = 2 * m
    mp = 2 * mp
    return Wigner_D(a, b, l, m, mp)


@nb.njit(nb.c16(nb.i8, nb.i8, nb.i8, nb.c16, nb.c16, nb.c16[:], nb.c16[:], nb.c16[:]),
         cache=True, fastmath=True, nogil=True)
def U_wD(l, mp, m, Ra, Rb, gradRa, gradRb, dU):
    '''Computes the hyperspherical harmonic function
    value for given Cayley-Klein parameters and angular
    momentum numbers.  The hyperspherical harmonics are the
    elements of the Wigner-D matrices.

    This function is an interface to the Wigner-D matrix
    function in angular_momentum.py.  The indexing and
    generation of the Cayley-Klein parameters is handled in
    the compute_uarray functions, which are handled by
    the SO4 bispectrum class.

    Parameters
    ----------
        l:  integer
        positive integer parameter, corresponds to 2 times the
        orbital angular momentum quantum number.

        mp:  integer
        integer parameter, corresponds to the magnetic quantum
        number by the relation m' = mp - l/2

        m:  integer
        integer parameter, corresponds to the magnetic quantum
        number by the relation m = m - l/2

        Ra:  complex
        Cayley-Klein parameter of unit quaternion.  Corresponds
        to the diagonal elements in the corresponding element of
        SU(2) in the 2x2 matrix representation.

        Rb:  complex
        Cayley-Klein parameter of unit quaternion.  Corresponds
        to the off-diagonal elements in the corresponding element of
        SU(2) in the 2x2 matrix representation.


    Returns
    -------
    complex, element of the Wigner-D matrix, see
    the corresponding function in angular_momentum.py
    '''
    m = m - l / 2
    mp = mp - l / 2
    m = 2 * m
    mp = 2 * mp
    return Wigner_D_wDerivative(Ra, Rb, l, m, mp, gradRa, gradRb, dU)


@nb.njit(nb.void(nb.f8, nb.f8, nb.f8, nb.f8, nb.f8, nb.i8, nb.c16[:, :], nb.i8[:]),
         cache=True, fastmath=True, nogil=True)
def compute_uarray_polynomial(x, y, z, psi, r, twol, ulist, idxu_block):
    '''Compute the Wigner-D matrix of order twol given an axis (x,y,z)
    and rotation angle 2*psi.

    This function constructs a unit quaternion representating a rotation
    of 2*psi through an axis defined by x,y,z; then populates an array of
    Wigner-D matrices of order twol for this rotation.  The Wigner-D matrices
    are calculated using a polynomial form.  See angular_momentum.Wigner_D for
    details.

    Parameters
    ----------
    x: float
    the x coordinate corresponding to the axis of rotation.

    y: float
    the y coordinate corresponding to the axis of rotation.

    z: float
    the z coordinate corresponding to the axis of rotation

    psi: float
    one half of the rotation angle

    r: float
    magnitude of the vector (x,y,z)

    twol: integer
    order of hyperspherical expansion

    ulist: 1-D complex array
    array to populate with D-matrix elements, mathematically
    this is a 3-D matrix, although we broadcast this to a 1-D
    matrix

    idxu_block: 1-D int array
    used to index ulist

    Returns
    -------
    None
    '''

    ldim = twol + 1

    # construct Cayley-Klein parameters for unit quaternion
    Ra = np.cos(psi) - 1j * np.sin(psi) / r * z
    Rb = np.sin(psi) / r * (y - x * 1j)

    # populate array with D matrix elements
    for l in range(0, ldim, 1):
        llu = idxu_block[l]
        for mb in range(0, l + 1, 1):
            for ma in range(0, l + 1, 1):
                ulist[llu, 0] = U(l, mb, ma, Ra, Rb)
                llu += 1
    return


@nb.njit(nb.void(nb.f8, nb.f8, nb.f8, nb.f8, nb.f8, nb.i8, nb.c16[:, :], nb.c16[:, :], nb.i8[:]),
         cache=True, fastmath=True, nogil=True)
def compute_uarray_polynomial_wD(x, y, z, psi, r, twol, ulist, dulist, idxu_block):
    '''Compute the Wigner-D matrix of order twol given an axis (x,y,z)
    and rotation angle 2*psi.

    This function constructs a unit quaternion representating a rotation
    of 2*psi through an axis defined by x,y,z; then populates an array of
    Wigner-D matrices of order twol for this rotation.  The Wigner-D matrices
    are calculated using a polynomial form.  See angular_momentum.Wigner_D for
    details.

    Parameters
    ----------
    x: float
    the x coordinate corresponding to the axis of rotation.

    y: float
    the y coordinate corresponding to the axis of rotation.

    z: float
    the z coordinate corresponding to the axis of rotation

    psi: float
    one half of the rotation angle

    r: float
    magnitude of the vector (x,y,z)

    twol: integer
    order of hyperspherical expansion

    ulist: 1-D complex array
    array to populate with D-matrix elements, mathematically
    this is a 3-D matrix, although we broadcast this to a 1-D
    matrix

    idxu_block: 1-D int array
    used to index ulist

    Returns
    -------
    None
    '''

    ldim = twol + 1

    # construct Cayley-Klein parameters for unit quaternion
    cospsi = np.cos(psi)
    sinpsi = np.sin(psi)
    gradr = np.array((x, y, z), np.complex128) / r
    Ra = cospsi - 1j * sinpsi / r * z
    Rb = sinpsi / r * (y - x * 1j)

    gradRa = -sinpsi * psi / r * gradr - 1j * z * cospsi / r / r * psi * gradr + 1j * z * sinpsi / r / r * gradr
    gradRa[2] += -1j * sinpsi / r

    gradRb = cospsi / r / r * psi * (y - x * 1j) * gradr - sinpsi / r / r * (y - x * 1j) * gradr

    gradRb[1] += sinpsi / r
    gradRb[0] += -1j * sinpsi / r

    # populate array with D matrix elements
    for l in range(0, ldim, 1):
        llu = idxu_block[l]
        for mb in range(0, l + 1, 1):
            for ma in range(0, l + 1, 1):
                ulist[llu, 0] = U_wD(l, mb, ma, Ra, Rb, gradRa, gradRb, dulist[llu, :])
                llu += 1

    return


@nb.njit(nb.void(nb.f8, nb.f8, nb.c16[:, :], nb.c16[:, :]),
         cache=True, fastmath=True, nogil=True)
def add_uarraytot(r, rcut, ulisttot, ulist):
    '''
    add the hyperspherical harmonic array for one neighbor to the
    expansion coefficient array
    '''
    sfac = compute_sfac(r, rcut)

    ulisttot += sfac * ulist
    return


@nb.njit(nb.void(nb.f8, nb.f8, nb.f8, nb.f8, nb.f8, nb.c16[:, :], nb.c16[:, :]),
         cache=True, fastmath=True, nogil=True)
def dudr(x, y, z, r, rcut, ulist, dulist):
    '''
    Compute the total derivative of the hyperspherical
    expansion coefficients.
    '''
    sfac = compute_sfac(r, rcut)
    gradr = np.zeros((len(ulist), 3), np.complex128)
    gradr[:, 0] = x
    gradr[:, 1] = y
    gradr[:, 2] = z
    gradr = gradr / r
    dsfac = compute_dsfac(r, rcut) * gradr

    dulist *= sfac
    dulist += ulist * dsfac
    return


@nb.njit(nb.void(nb.i8, nb.i8[:, :], nb.f8[:], nb.i8[:, :, :], nb.i8[:], nb.c16[:, :], nb.c16[:, :]),
         cache=True, fastmath=True, nogil=True)
def compute_zi(idxz_max, idxz, cglist, idxcg_block, idxu_block, ulisttot, zlist):
    '''
    Precompute the Kronoecker product of two rotated expansion coefficient tensors
    using Clebsch-Gordan expansion
    '''

    for llz in range(0, idxz_max, 1):
        l1 = idxz[llz, 0]
        l2 = idxz[llz, 1]
        l = idxz[llz, 2]
        ma1min = idxz[llz, 3]
        ma2max = idxz[llz, 4]
        na = idxz[llz, 5]
        mb1min = idxz[llz, 6]
        mb2max = idxz[llz, 7]
        nb = idxz[llz, 8]

        cgblock = cglist[idxcg_block[l1, l2, l]::]

        zlist[llz] = 0

        llu1 = idxu_block[l1] + (l1 + 1) * mb1min
        llu2 = idxu_block[l2] + (l2 + 1) * mb2max
        icgb = mb1min * (l2 + 1) + mb2max

        for ib in range(0, nb, 1):
            suma1 = 0

            u1 = ulisttot[llu1::]
            u2 = ulisttot[llu2::]

            ma1 = ma1min
            ma2 = ma2max
            icga = ma1min * (l2 + 1) + ma2max
            for ia in range(0, na, 1):
                suma1 += cgblock[icga] * u1[ma1, 0] * u2[ma2, 0]
                ma1 += 1
                ma2 -= 1
                icga += l2

            zlist[llz] += cgblock[icgb] * suma1

            llu1 += l1 + 1
            llu2 -= l2 + 1
            icgb += l2

    return


@nb.njit(nb.void(nb.i8, nb.i8[:, :], nb.i8[:, :, :], nb.i8[:], nb.c16[:, :], nb.c16[:, :], nb.c16[:]),
         cache=True, fastmath=True, nogil=True)
def compute_bi(ncoefs, idxb, idxz_block, idxu_block, ulisttot, zlist, blist):
    '''
    compute the bispectrum components from the Kronoecker product of the hermitian adjoint
    of the expansion coefficient array with the Z list (see compute Z_i for description)
    '''

    for llb in range(0, ncoefs, 1):
        l1 = idxb[llb, 0]
        l2 = idxb[llb, 1]
        l = idxb[llb, 2]

        llz = idxz_block[l1][l2][l]
        llu = idxu_block[l]

        sumzu = 0

        mb = 0
        while 2 * mb < l:
            for ma in range(0, l + 1, 1):
                sumzu += ulisttot[llu, 0].conjugate() * zlist[llz, 0]
                llz += 1
                llu += 1

            mb += 1

            # for l even handle middle column in a different manner

        if l % 2 == 0:
            mb = l / 2
            for ma in range(0, mb, 1):
                sumzu += ulisttot[llu, 0].conjugate() * zlist[llz, 0]
                llz += 1
                llu += 1

            sumzu += 0.5 * (ulisttot[llu, 0].conjugate() * zlist[llz, 0])

        blist[llb] = 2.0 * sumzu

    return


@nb.njit(nb.void(nb.i8, nb.i8[:, :], nb.i8[:, :, :], nb.i8[:], nb.c16[:, :], nb.c16[:, :], nb.c16[:, :]),
         cache=True, fastmath=True, nogil=True)
def compute_dbidrj(ncoefs, idxb, idxz_block, idxu_block, dulist, zlist, dblist):
    '''
    Compute the gradient of the bispectrum components
    '''

    for llb in range(0, ncoefs, 1):
        l1 = idxb[llb, 0]
        l2 = idxb[llb, 1]
        l = idxb[llb, 2]

        llz = idxz_block[l1, l2, l]
        llu = idxu_block[l]

        sumzdu = np.zeros(3, np.complex128)

        mb = 0
        while 2 * mb < l:
            for ma in range(0, l + 1, 1):
                dudr = np.conj(dulist[llu])
                z = zlist[llz]
                sumzdu += dudr * z
                llu += 1
                llz += 1
            mb += 1

        if l % 2 == 0:
            mb = l / 2
            for ma in range(0, mb, 1):
                dudr = np.conj(dulist[llu])
                z = zlist[llz]
                sumzdu += dudr * z
                llu += 1
                llz += 1

            ma = mb
            dudr = np.conj(dulist[llu])
            z = zlist[llz]
            sumzdu += dudr * z * 0.5
            llu += 1
            llz += 1

        dblist[llb] += 2.0 * sumzdu

        l1fac = (l + 1) / (l1 + 1.0)

        llz = idxz_block[l, l2, l1]
        llu = idxu_block[l1]

        sumzdu = np.zeros(3, np.complex128)

        mb = 0
        while 2 * mb < l1:
            for ma in range(0, l1 + 1, 1):
                dudr = np.conj(dulist[llu])
                z = zlist[llz]
                sumzdu += dudr * z
                llu += 1
                llz += 1
            mb += 1

        if l1 % 2 == 0:
            mb = l1 / 2
            for ma in range(0, mb, 1):
                dudr = np.conj(dulist[llu])
                z = zlist[llz]
                sumzdu += dudr * z
                llu += 1
                llz += 1

            ma = mb
            dudr = np.conj(dulist[llu])
            z = zlist[llz]
            sumzdu += dudr * z * 0.5
            llu += 1
            llz += 1

        dblist[llb] += 2.0 * sumzdu * l1fac

        l2fac = (l + 1) / (l2 + 1.0)

        llz = idxz_block[l, l1, l2]
        llu = idxu_block[l2]

        sumzdu = np.zeros(3, np.complex128)

        mb = 0
        while 2 * mb < l2:
            for ma in range(0, l2 + 1, 1):
                dudr = np.conj(dulist[llu])
                z = zlist[llz]
                sumzdu += dudr * z
                llu += 1
                llz += 1
            mb += 1

        if l2 % 2 == 0:
            mb = l2 / 2
            for ma in range(0, mb, 1):
                dudr = np.conj(dulist[llu])
                z = zlist[llz]
                sumzdu += dudr * z
                llu += 1
                llz += 1

            ma = mb
            dudr = np.conj(dulist[llu])
            z = zlist[llz]
            sumzdu += dudr * z * 0.5
            llu += 1
            llz += 1

        dblist[llb] += 2.0 * sumzdu * l2fac


@nb.njit(nb.void(nb.c16[:, :]),
         cache=True, fastmath=True, nogil=True)
def zero_1d(arr):
    # zero a generic 1-d array
    for i in range(arr.shape[0]):
        arr[i] = 0
    return


@nb.njit(nb.void(nb.c16[:, :]),
         cache=True, fastmath=True, nogil=True)
def zero_2d(arr):
    # zero a generic 2d array
    for i in range(arr.shape[0]):
        for j in range(arr.shape[1]):
            arr[i, j] = 0
    return


@nb.njit(nb.void(nb.c16[:, :, :]),
         cache=True, fastmath=True, nogil=True)
def zero_3d(arr):
    # zero a generic 3d array
    for i in range(arr.shape[0]):
        for j in range(arr.shape[1]):
            for k in range(arr.shape[2]):
                arr[i, j, k] = 0
    return


@nb.njit(nb.void(nb.f8[:, :], nb.f8[:, :, :], nb.i8[:, :], nb.i8[:, :], nb.i8[:], nb.i8, nb.f8,
                 nb.b1, nb.b1, nb.b1, nb.c16[:, :], nb.c16[:, :, :], nb.c16[:, :, :, :]),
         cache=True, fastmath=True, nogil=True)
def get_bispectrum_components(center_atoms, neighborlist, seq, neighbor_ANs, site_ANs,
                              twolmax, rcut, norm, derivative, stress, blist, dblist, bstress):
    '''
    Calculate the bispectrum components, and their derivatives (if specified)
    for a given neighbor list.  This is the main work function.
    '''
    # build index lists, normalization array, and CG array
    ldim = twolmax + 1

    # array for indexing Clebsch-Gordan coefficients
    idxcg_block = np.zeros((ldim, ldim, ldim), dtype=np.int64)

    # variable to count the number of CG coefficients for a given twol
    idxcg_count = 0
    # populate the index list and count the number of CG coefficients
    for l1 in range(0, ldim, 1):
        for l2 in range(0, l1 + 1, 1):
            for l in range(l1 - l2, min(twolmax, l1 + l2) + 1, 2):
                idxcg_block[l1, l2, l] = idxcg_count
                for m1 in range(0, l1 + 1, 1):
                    for m2 in range(0, l2 + 1, 1):
                        idxcg_count += 1

    # allocate array for Clebsch-Gordan coefficients
    # then populate array
    cglist = np.zeros(idxcg_count, dtype=np.float64)
    init_clebsch_gordan(twolmax, cglist)

    # index list for u array
    idxu_block = np.zeros(ldim, dtype=np.int64)

    # populate the index list for u arrays and count
    # the number of u arrays
    idxu_count = 0
    for l in range(0, ldim, 1):
        idxu_block[l] = idxu_count
        for mb in range(0, l + 1, 1):
            for ma in range(0, l + 1, 1):
                idxu_count += 1

    # populate array for normalization of Wigner-D functions
    u_norm = np.ones((idxu_count, 1), dtype=np.float64)

    # if normalization option is enabled populate array of
    # normalization factors
    if norm == True:
        idxu_count = 0
        for l in range(0, ldim, 1):
            idxu_block[l] = idxu_count
            for mb in range(0, l + 1, 1):
                for ma in range(0, l + 1, 1):
                    u_norm[idxu_count] = np.sqrt(l + 1) / 4 / np.pi
                    idxu_count += 1

    # index list for B
    ncoefs = blist.shape[1]
    idxb = np.zeros((ncoefs, 3), dtype=np.int64)

    # populate index list with bispectrum component indices
    idxb_count = 0
    for l1 in range(0, ldim, 1):
        for l2 in range(0, l1 + 1, 1):
            for l in range(l1 - l2, min(twolmax, l1 + l2) + 1, 2):
                if l >= l1:
                    idxb[idxb_count, 0] = l1
                    idxb[idxb_count, 1] = l2
                    idxb[idxb_count, 2] = l
                    idxb_count += 1

    # index list for zlist
    idxz_count = 0
    for l1 in range(0, ldim, 1):
        for l2 in range(0, l1 + 1, 1):
            for l in range(l1 - l2, min(twolmax, l1 + l2) + 1, 2):
                mb = 0
                while 2 * mb <= l:
                    for ma in range(0, l + 1, 1):
                        idxz_count += 1
                    mb += 1

    idxz = np.zeros((idxz_count, 10), dtype=np.int64)

    idxz_block = np.zeros((ldim, ldim, ldim), dtype=np.int64)

    idxz_count = 0
    for l1 in range(0, ldim, 1):
        for l2 in range(0, l1 + 1, 1):
            for l in range(l1 - l2, min(twolmax, l1 + l2) + 1, 2):
                idxz_block[l1, l2, l] = idxz_count

                # find right beta [llb] entry
                # multiply and divide by l+1 factors
                # account for multiplicity of 1, 2, or 3

                mb = 0
                while 2 * mb <= l:
                    for ma in range(0, l + 1, 1):
                        idxz[idxz_count, 0] = l1  # l1 at 0 position
                        idxz[idxz_count, 1] = l2  # l2 at 1 position
                        idxz[idxz_count, 2] = l  # l at 2 position

                        ma1min = max(0, (2 * ma - l - l2 + l1) / 2)
                        # ma1min at 3 position
                        idxz[idxz_count, 3] = ma1min

                        ma2max = (2 * ma - l -
                                  (2 * idxz[idxz_count, 3] - l1) + l2) / 2
                        # ma2max at 4 position
                        idxz[idxz_count, 4] = ma2max

                        na = min(l1, (2 * ma - l + l2 + l1) / 2) - \
                             idxz[idxz_count, 3] + 1
                        idxz[idxz_count, 5] = na  # na at 5 position

                        mb1min = max(0, (2 * mb - l - l2 + l1) / 2)
                        # mb1min at 6 position
                        idxz[idxz_count, 6] = mb1min

                        mb2max = (2 * mb - l -
                                  (2 * idxz[idxz_count, 6] - l1) + l2) / 2
                        # mb2max at 7 position
                        idxz[idxz_count, 7] = mb2max

                        nb = min(l1, (2 * mb - l + l2 + l1) / 2) - \
                             idxz[idxz_count, 6] + 1
                        idxz[idxz_count, 8] = nb  # nb at 8 position

                        # apply to z(l1,l2,jma,mb) to unique element of
                        # y(l)

                        llu = idxu_block[l] + (l + 1) * mb + ma
                        idxz[idxz_count, 9] = llu  # llu at 9 position

                        idxz_count += 1

                    mb += 1

    # calculate bispectrum components
    npairs = neighborlist.shape[0]
    nneighbors = neighborlist.shape[1]

    ulisttot = np.zeros((idxu_count, 1), dtype=np.complex128)
    zlist = np.zeros((idxz_count, 1), dtype=np.complex128)
    dulist = np.zeros((nneighbors, idxu_count, 3), dtype=np.complex128)
    ulist = np.zeros((idxu_count, 1), dtype=np.complex128)
    tempdb = np.zeros((idxb_count, 3), dtype=np.complex128)
    Rj = np.zeros(3, dtype=np.float64)

    if derivative == True:
        if stress == True:
            isite = seq[0, 0]
            nstart = 0
            nsite = 0  # 0,0
            for n in range(npairs):
                i, j = seq[n]
                # talk to Qiang about potential issue here
                # when there are neighborlists where there
                # are no i-i atom pairs
                if i == j:
                    nsite = n
                weight = neighbor_ANs[n, 0]

                if i != isite:
                    zero_1d(ulist)
                    addself_uarraytot(twolmax, idxu_block, ulist)
                    ulist *= site_ANs[isite]
                    ulisttot += ulist
                    ulisttot *= u_norm
                    compute_zi(idxz_count, idxz, cglist, idxcg_block, idxu_block,
                               ulisttot, zlist)
                    compute_bi(ncoefs, idxb, idxz_block, idxu_block, ulisttot, zlist,
                               blist[isite])
                    for N in range(nstart, n, 1):
                        I, J = seq[N]
                        Ri = center_atoms[I]
                        Weight = neighbor_ANs[N, 0]
                        zero_3d(dulist)
                        for neighbor in prange(nneighbors):
                            x = neighborlist[N, neighbor, 0]
                            y = neighborlist[N, neighbor, 1]
                            z = neighborlist[N, neighbor, 2]
                            r = np.sqrt(x * x + y * y + z * z)
                            if r < 10 ** (-8):
                                continue
                            psi = np.pi * r / rcut
                            zero_1d(ulist)
                            compute_uarray_polynomial_wD(x, y, z, psi, r, twolmax, ulist,
                                                         dulist[neighbor], idxu_block)

                            dudr(x, y, z, r, rcut, ulist, dulist[neighbor])

                            dulist[neighbor] *= Weight
                            dulist[neighbor] *= u_norm

                            zero_2d(tempdb)

                            compute_dbidrj(ncoefs, idxb, idxz_block, idxu_block, dulist[neighbor],
                                           zlist, tempdb)

                            dblist[N] += tempdb
                            if I != J:
                                dblist[nsite] -= tempdb

                            Rj[0] = x + Ri[0]
                            Rj[1] = y + Ri[1]
                            Rj[2] = z + Ri[2]

                            for k in range(idxb_count):
                                bstress[nsite, k] += np.outer(Ri, tempdb[k])
                                bstress[N, k] -= np.outer(Rj, tempdb[k])

                    isite = i
                    nstart = n
                    zero_1d(ulisttot)
                    zero_1d(zlist)
                    zero_3d(dulist)
                # end if i != isite

                for neighbor in prange(nneighbors):
                    # get components of separation vector and its magnitude
                    # this is also the axis of rotation
                    x = neighborlist[n, neighbor, 0]
                    y = neighborlist[n, neighbor, 1]
                    z = neighborlist[n, neighbor, 2]
                    r = np.sqrt(x * x + y * y + z * z)
                    if r < 10 ** (-8):
                        continue
                    # angle of rotation
                    psi = np.pi * r / rcut
                    # populate ulist and dulist with Wigner U functions
                    # and derivatives
                    zero_1d(ulist)
                    compute_uarray_polynomial(x, y, z, psi, r, twolmax, ulist,
                                              idxu_block)

                    ulist *= weight

                    add_uarraytot(r, rcut, ulisttot, ulist)

            zero_1d(ulist)
            addself_uarraytot(twolmax, idxu_block, ulist)
            ulist *= site_ANs[isite]
            ulisttot += ulist
            ulisttot *= u_norm
            compute_zi(idxz_count, idxz, cglist, idxcg_block, idxu_block,
                       ulisttot, zlist)
            compute_bi(ncoefs, idxb, idxz_block, idxu_block, ulisttot, zlist,
                       blist[isite])
            for N in range(nstart, npairs, 1):
                I, J = seq[N]
                Ri = center_atoms[I]
                Weight = neighbor_ANs[N, 0]
                zero_3d(dulist)
                for neighbor in prange(nneighbors):
                    x = neighborlist[N, neighbor, 0]
                    y = neighborlist[N, neighbor, 1]
                    z = neighborlist[N, neighbor, 2]
                    r = np.sqrt(x * x + y * y + z * z)
                    if r < 10 ** (-8):
                        continue
                    psi = np.pi * r / rcut
                    zero_1d(ulist)
                    compute_uarray_polynomial_wD(x, y, z, psi, r, twolmax, ulist,
                                                 dulist[neighbor], idxu_block)

                    dudr(x, y, z, r, rcut, ulist, dulist[neighbor])

                    dulist[neighbor] *= Weight
                    dulist[neighbor] *= u_norm

                    zero_2d(tempdb)

                    compute_dbidrj(ncoefs, idxb, idxz_block, idxu_block, dulist[neighbor],
                                   zlist, tempdb)

                    dblist[N] += tempdb
                    if I != J:
                        dblist[nsite] -= tempdb

                    Rj[0] = x + Ri[0]
                    Rj[1] = y + Ri[1]
                    Rj[2] = z + Ri[2]

                    for k in range(idxb_count):
                        bstress[nsite, k] += np.outer(Ri, tempdb[k])
                        bstress[N, k] -= np.outer(Rj, tempdb[k])


        else:
            isite = seq[0, 0]
            nstart = 0
            nsite = 0  # 0,0
            for n in range(npairs):
                i, j = seq[n]
                if i == j:
                    nsite = n
                weight = neighbor_ANs[n, 0]

                if i != isite:
                    zero_1d(ulist)
                    addself_uarraytot(twolmax, idxu_block, ulist)
                    ulist *= site_ANs[isite]
                    ulisttot += ulist
                    ulisttot *= u_norm
                    compute_zi(idxz_count, idxz, cglist, idxcg_block, idxu_block,
                               ulisttot, zlist)
                    compute_bi(ncoefs, idxb, idxz_block, idxu_block, ulisttot, zlist,
                               blist[isite])
                    for N in range(nstart, n, 1):
                        I, J = seq[N]
                        Weight = neighbor_ANs[N, 0]
                        zero_3d(dulist)
                        for neighbor in prange(nneighbors):
                            x = neighborlist[N, neighbor, 0]
                            y = neighborlist[N, neighbor, 1]
                            z = neighborlist[N, neighbor, 2]
                            r = np.sqrt(x * x + y * y + z * z)
                            if r < 10 ** (-8):
                                continue
                            psi = np.pi * r / rcut
                            zero_1d(ulist)
                            compute_uarray_polynomial_wD(x, y, z, psi, r, twolmax, ulist,
                                                         dulist[neighbor], idxu_block)

                            dudr(x, y, z, r, rcut, ulist, dulist[neighbor])

                            dulist[neighbor] *= Weight
                            dulist[neighbor] *= u_norm

                            zero_2d(tempdb)

                            compute_dbidrj(ncoefs, idxb, idxz_block, idxu_block, dulist[neighbor],
                                           zlist, tempdb)

                            dblist[N] += tempdb
                            if I != J:
                                dblist[nsite] -= tempdb

                    isite = i
                    nstart = n
                    zero_1d(ulisttot)
                    zero_1d(zlist)
                    zero_3d(dulist)
                # end if i != isite

                for neighbor in prange(nneighbors):
                    # get components of separation vector and its magnitude
                    # this is also the axis of rotation
                    x = neighborlist[n, neighbor, 0]
                    y = neighborlist[n, neighbor, 1]
                    z = neighborlist[n, neighbor, 2]
                    r = np.sqrt(x * x + y * y + z * z)
                    if r < 10 ** (-8):
                        continue
                    # angle of rotation
                    psi = np.pi * r / rcut
                    # populate ulist and dulist with Wigner U functions
                    # and derivatives
                    zero_1d(ulist)
                    compute_uarray_polynomial(x, y, z, psi, r, twolmax, ulist,
                                              idxu_block)

                    ulist *= weight

                    add_uarraytot(r, rcut, ulisttot, ulist)

            zero_1d(ulist)
            addself_uarraytot(twolmax, idxu_block, ulist)
            ulist *= site_ANs[isite]
            ulisttot += ulist
            ulisttot *= u_norm
            compute_zi(idxz_count, idxz, cglist, idxcg_block, idxu_block,
                       ulisttot, zlist)
            compute_bi(ncoefs, idxb, idxz_block, idxu_block, ulisttot, zlist,
                       blist[isite])
            for N in range(nstart, npairs, 1):
                I, J = seq[N]
                Weight = neighbor_ANs[N, 0]
                zero_3d(dulist)
                for neighbor in prange(nneighbors):
                    x = neighborlist[N, neighbor, 0]
                    y = neighborlist[N, neighbor, 1]
                    z = neighborlist[N, neighbor, 2]
                    r = np.sqrt(x * x + y * y + z * z)
                    if r < 10 ** (-8):
                        continue
                    psi = np.pi * r / rcut
                    zero_1d(ulist)
                    compute_uarray_polynomial_wD(x, y, z, psi, r, twolmax, ulist,
                                                 dulist[neighbor], idxu_block)

                    dudr(x, y, z, r, rcut, ulist, dulist[neighbor])

                    dulist[neighbor] *= Weight
                    dulist[neighbor] *= u_norm

                    zero_2d(tempdb)

                    compute_dbidrj(ncoefs, idxb, idxz_block, idxu_block, dulist[neighbor],
                                   zlist, tempdb)

                    dblist[N] += tempdb
                    if I != J:
                        dblist[nsite] -= tempdb


    else:
        isite = seq[0, 0]
        nstart = 0
        nsite = 0  # 0,0
        for n in range(npairs):
            i, j = seq[n]
            if i == j:
                nsite = n
            weight = neighbor_ANs[n, 0]

            if i != isite:
                zero_1d(ulist)
                addself_uarraytot(twolmax, idxu_block, ulist)
                ulist *= site_ANs[isite]
                ulisttot += ulist
                ulisttot *= u_norm
                compute_zi(idxz_count, idxz, cglist, idxcg_block, idxu_block,
                           ulisttot, zlist)
                compute_bi(ncoefs, idxb, idxz_block, idxu_block, ulisttot, zlist,
                           blist[isite])

                isite = i
                nstart = n
                zero_1d(ulisttot)
                zero_1d(zlist)
            # end if i != isite

            for neighbor in prange(nneighbors):
                # get components of separation vector and its magnitude
                # this is also the axis of rotation
                x = neighborlist[n, neighbor, 0]
                y = neighborlist[n, neighbor, 1]
                z = neighborlist[n, neighbor, 2]
                r = np.sqrt(x * x + y * y + z * z)
                if r < 10 ** (-8):
                    continue
                # angle of rotation
                psi = np.pi * r / rcut
                # populate ulist and dulist with Wigner U functions
                # and derivatives
                zero_1d(ulist)
                compute_uarray_polynomial(x, y, z, psi, r, twolmax, ulist,
                                          idxu_block)

                ulist *= weight

                add_uarraytot(r, rcut, ulisttot, ulist)

        zero_1d(ulist)
        addself_uarraytot(twolmax, idxu_block, ulist)
        ulist *= site_ANs[isite]
        ulisttot += ulist
        ulisttot *= u_norm
        compute_zi(idxz_count, idxz, cglist, idxcg_block, idxu_block,
                   ulisttot, zlist)
        compute_bi(ncoefs, idxb, idxz_block, idxu_block, ulisttot, zlist,
                   blist[isite])

    return


if __name__ == "__main__":
    from ase.io import read

    # ---------------------- Options ------------------------
    parser = OptionParser()
    parser.add_option("-c", "--crystal", dest="structure",
                      help="crystal from file, cif or poscar, REQUIRED",
                      metavar="crystal")

    parser.add_option("-r", "--rcut", dest="rcut", default=4.0, type=float,
                      help="cutoff for neighbor calcs, default: 4.0"
                      )

    parser.add_option("-l", "--lmax", dest="lmax", default=1, type=int,
                      help="lmax, default: 1"
                      )

    parser.add_option("-s", dest="stress", default=True,
                      action='store_true', help='derivative flag')

    parser.add_option("-f", dest="der", default=True,
                      action='store_false', help='derivative flag')

    (options, args) = parser.parse_args()

    if options.structure is None:
        from ase.build import bulk

        test = bulk('Si', 'diamond', a=5.459)
        cell = test.get_cell()
        cell[0, 1] += 0.5
        test.set_cell(cell)
    else:
        test = read(options.structure, format='vasp')
    print(test)
    lmax = options.lmax
    rcut = options.rcut
    der = options.der
    stress = options.stress

    # import time
    f = SO4_Bispectrum(lmax, rcut, derivative=False, stress=False, normalize_U=False)
    x = f.calculate(test)
    # start2 = time.time()
    # for key, item in x.items():
    #    print(key, item)
    # print('time elapsed: {}'.format(start2 - start1))
    # print(x['rdxdr'].shape)
    # print(x['rdxdr'])
    # print(np.einsum('ijklm->klm', x['rdxdr']))
    print(x['x'])



from typing import List

from featurebox.featurizers.atom.mapper import AtomPymatgenPropMap
from featurebox.featurizers.base_transform import BaseFeature
from featurebox.featurizers.state.state_mapper import StructurePymatgenPropMap
from featurebox.featurizers.state.statistics import WeightedAverage


class BatchFeature(BaseFeature):
    """Script for generate batch_data, could be copied and user-defined."""

    def __init__(self, data_type: str = "compositions", user_convert: BaseFeature = None, n_jobs: int = 1,
                 on_errors: str = 'raise', return_type: str = 'any',
                 batch_calculate: bool = False,
                 batch_size: int = 30):
        """

        Parameters
        ----------
        data_type:str
            Predefined name ["elements", "compositions", "structures"]
        user_convert:BatchFeature
            which contain `convert` method.
        """
        super(BatchFeature, self).__init__(n_jobs, on_errors=on_errors, return_type=return_type,
                                           batch_calculate=batch_calculate, batch_size=batch_size)

        self.structure_c = StructurePymatgenPropMap(prop_name=["density",
                                                               "volume",
                                                               "ntypesp",

                                                               "_lattice.lengths",
                                                               "_lattice.angles",

                                                               "composition.average_electroneg",
                                                               "composition.num_atoms",
                                                               "composition.total_electrons",
                                                               "composition.weight",

                                                               ], n_jobs=1)

        func_map = [
            "atomic_radius",
            "atomic_mass",
            "Z",
            "X",
            "number",
            "max_oxidation_state",
            "min_oxidation_state",
            "row",
            "group",
            "atomic_radius_calculated",

            "van_der_waals_radius",
            "mendeleev_no",
            "molar_volume",
            "boiling_point",
            "melting_point",
            "critical_temperature",
            "density_of_solid",
            "average_ionic_radius",
            "average_cationic_radius",
            "average_anionic_radius",
        ]
        appm = AtomPymatgenPropMap(prop_name=func_map, n_jobs=1, search_tp="name")
        self.composition_c = WeightedAverage(appm, n_jobs=1, return_type="df")
        self.composition_c.set_feature_labels(func_map)

        self.element_c = AtomPymatgenPropMap(prop_name=func_map, n_jobs=1,
                                             search_tp="name",
                                             return_type="df")

        if user_convert is not None:
            self.data_type = "user-defined"
            self.convert_method = user_convert
        else:
            self.data_type = data_type
            self.maps = {
                "elements": self.element_c,
                "compositions": self.composition_c,
                "structures": self.structure_c,
                # "electron":self.electron_c,
                # "dos":self.dos_dict,
                # "band_gap":self.band_gap_dict,
            }
            if self.data_type not in self.maps:
                raise TypeError("Undefined name for data_type, if use your self convert, please set 'user_convert'.")
            self.convert_method = self.maps[self.data_type]

    def convert(self, d):
        if self.data_type == "compositions":
            return self.convert_method.convert(d)
        elif self.data_type == "elements":
            return self.convert_method.convert(d)
        elif self.data_type == "structures":
            return self.convert_method.convert(d)

    @property
    def feature_labels(self):
        return self.convert_method.feature_labels

    def set_feature_labels(self, values: List[str]):
        self.convert_method.set_feature_labels(values)



# -*- coding: utf-8 -*-

# @Time   : 2019/5/26 0:47
# @Author : Administrator
# @Project : feature_preparation
# @FileName: mutibase.py
# @Software: PyCharm
from typing import List

import numpy as np


class MutiBase(object):
    """Base method for binding"""

    def __init__(self, muti_grade: int = 2, muti_index: List = None, must_index: List = None):
        """

        Parameters
        ----------
        muti_grade:
            binding_group size, calculate the correction between binding
        muti_index:list
            the range of muti_grade:[min,max)
        must_index:list,None
            the columns force to index
        """
        self.muti_grade = muti_grade
        self.muti_index = muti_index
        self.must_index = must_index

    @property
    def check_muti(self):
        muti_index = self.muti_index
        if muti_index is None:
            return False
        elif isinstance(muti_index, (list, tuple)):
            if len(muti_index) == 2 and isinstance(muti_index[0], int) and isinstance(muti_index[1], int):
                return tuple(range(*muti_index))
        else:
            raise TypeError("muti_index should be None or iterable type with 2 number")

    @property
    def check_must(self):
        must_index = self.must_index
        if must_index is None:
            return False
        elif isinstance(must_index, (list, tuple)):
            if all([isinstance(_, int) for _ in must_index]) and 0 < len(must_index) <= 2:
                if len(must_index) == 1:
                    must_slice = tuple(must_index)
                else:
                    must_slice = tuple(range(*must_index))
                return must_slice
        else:
            raise TypeError("must_index should be None or iterable type with less than 2 number")

    def feature_fold(self, feature):
        muti_grade, muti_index = self.muti_grade, self.muti_index
        if self.check_muti:
            feature = np.sort(feature)
            single = np.array([_ for _ in feature if _ < muti_index[0] or _ >= muti_index[1]])
            com_com = np.array([_ for _ in feature if muti_index[1] > _ >= muti_index[0]])
            com_sin = com_com[::muti_grade]
            return np.sort(np.hstack((single, com_sin))).astype(int)
        else:
            return np.sort(feature).astype(int)

    def feature_unfold(self, feature):
        muti_grade, muti_index = self.muti_grade, self.muti_index
        if self.check_muti:
            single = np.array([_ for _ in feature if _ < muti_index[0] or _ >= muti_index[1]])
            com_sin = np.array([_ for _ in feature if muti_index[1] > _ >= muti_index[0]])
            com_com = list(com_sin)
            while muti_grade - 1:
                com_com.extend(com_sin + (muti_grade - 1))
                muti_grade -= 1
            return np.sort(list(set(np.hstack((single, np.array(com_com)))))).astype(int)
        else:
            return np.sort(np.array(feature)).astype(int)

    def feature_must_fold(self, feature):
        must_index = self.must_index
        if must_index:
            feature = list(feature)
            if len(must_index) == 1 and must_index[0] not in feature:
                feature.append(must_index[0])
            else:
                must_feature = list(range(*must_index))
                must_feature = self.feature_fold(must_feature)
                feature.extend([j for j in must_feature if j not in feature])
            return np.sort(feature).astype(int)
        else:
            return np.sort(feature).astype(int)

    def feature_must_unfold(self, feature):
        must_index = self.must_index
        if must_index:
            feature = list(feature)
            if len(must_index) == 1 and must_index[0] not in feature:
                feature.append(must_index[0])
            else:
                must_feature = list(range(*must_index))
                feature.extend([j for j in must_feature if j not in feature])
            return np.sort(feature).astype(int)
        else:
            return np.sort(feature).astype(int)



from sklearn.datasets import load_boston
from sklearn.svm import SVR

from featurebox.selection.backforward import BackForward

X, y = load_boston(return_X_y=True)
svr = SVR()
bf = BackForward(svr, primary_feature=4, random_state=1, refit=True)
new_x = bf.fit_transform(X[:50], y[:50])
test_score = bf.score(X[50:], y[50:])


"""
This part is used to check the element.
"""
from typing import List, Callable

import numpy as np

ALL_ELE_NAME = ("H", "He", "Li", "Be", "B", "C", "N", "O", "F", "Ne", "Na", "Mg", "Al", "Si",
                "P", "S", "Cl", "Ar", "K", "Ca", "Sc", "Ti", "V", "Cr", "Mn", "Fe", "Co", "Ni",
                "Cu", "Zn", "Ga", "Ge", "As", "Se", "Br", "Kr", "Rb", "Sr", "Y", "Zr", "Nb", "Mo",
                "Tc", "Ru", "Rh", "Pd", "Ag", "Cd", "In", "Sn", "Sb", "Te", "I", "Xe", "Cs", "Ba",
                "La", "Ce", "Pr", "Nd", "Pm", "Sm", "Eu", "Gd", "Tb", "Dy", "Ho", "Er", "Tm", "Yb",
                "Lu", "Hf", "Ta", "W", "Re", "Os", "Ir", "Pt", "Au", "Hg", "Tl", "Pb", "Bi", "Po",
                "At", "Rn", "Fr", "Ra", "Ac", "Th", "Pa", "U", "Np", "Pu", "Am", "Cm", "Bk", "Cf",
                "Es", "Fm", "Md", "No", "Lr", "Rf", "Db", "Sg", "Bh", "Hs", "Mt", "Ds", "Rg", "Cn",
                "Nh", "Fl", "Mc", "Lv", "Ts", "Og")  # 118

ALL_N_ELE_MAP = {v + 1: k for k, v in zip(ALL_ELE_NAME, range(len(ALL_ELE_NAME)))}
ALL_ELE_N_MAP = {k: v + 1 for k, v in zip(ALL_ELE_NAME, range(len(ALL_ELE_NAME)))}

AVAILABLE_ELE_NAME = ('H', 'He', 'Li', 'Be', 'B', 'C', 'N', 'O', 'F', 'Ne', 'Na', 'Mg', 'Al', 'Si', 'P', 'S',
                      'Cl', 'Ar', 'K', 'Ca', 'Sc', 'Ti', 'V', 'Cr', 'Mn', 'Fe', 'Co', 'Ni', 'Cu', 'Zn', 'Ga',
                      'Ge', 'As', 'Se', 'Br', 'Kr', 'Rb', 'Sr', 'Y', 'Zr', 'Nb', 'Mo', 'Tc', 'Ru', 'Rh', 'Pd',
                      'Ag', 'Cd', 'In', 'Sn', 'Sb', 'Te', 'I', 'Xe', 'Cs', 'Ba', 'La', 'Ce', 'Pr', 'Nd', 'Pm',
                      'Sm', 'Eu', 'Gd', 'Tb', 'Dy', 'Ho', 'Er', 'Tm', 'Yb', 'Lu', 'Hf', 'Ta', 'W', 'Re', 'Os',
                      'Ir', 'Pt', 'Au', 'Hg', 'Tl', 'Pb', 'Bi',
                      # ######### 'Po', 'At', 'Rn', 'Fr', 'Ra', # 84-88
                      'Ac', 'Th', 'Pa', 'U',
                      # ########## "Np", "Pu", "Am", "Cm", "Bk", "Cf", "Es", "Fm", "Md", "No", "Lr", "Rf", "Db", "Sg",
                      # ########## "Bh", "Hs", "Mt", "Ds", "Rg", "Cn", "Nh", "Fl", "Mc", "Lv", "Ts", "Og"
                      )

AVAILABLE_ELE_NUMBER = tuple(list((range(1, 84))) + [89, 90, 91, 92])


class CheckElements:
    """Check the element in preprocessing."""

    def __init__(self, check_method: str = "name", func: Callable = lambda x: x):
        """

        Parameters
        ----------
        check_method: str
            Check by number or name of element.
        func: callable
            Processing for elements.
            such as for element in pymatgen:

            >>> func = lambda x: [x.Z, ]
            >>> func2 = lambda x: [x.name, ]

        Examples
        ---------
        >>> ce = CheckElements.from_list(check_method="name",grouped=False)
        >>> ce.check(["Na","Al","Ta"])
        ['Na', 'Al', 'Ta']
        >>> ce = CheckElements.from_list(check_method="name",grouped=True)
        >>> ce.check([["Na","Al"],["Na","Ta"]])
        [['Na', 'Al'], ['Na', 'Ta']]
        >>> ce.check([["Na","Al"],["Na","Ra"],["Zn","H"]])
        The 1 (st,ed,th) sample ['Na', 'Ra'] is with element out of AVAILABLE_ELE_NAME
         please to check_data.py for more information.
        [['Na', 'Al'], ['Zn', 'H']]
        >>> ce.passed_idx()
        array([0, 2], dtype=int64)

        Examples
        ---------
        >>> ce = CheckElements.from_pymatgen_structures()
        ...

        """

        if check_method == "name":
            check_method = AVAILABLE_ELE_NAME
        elif check_method == "number":
            check_method = AVAILABLE_ELE_NUMBER
        else:
            raise TypeError("check_method='name' or 'name'")
        self.check_method = check_method
        self.func = func
        self.mark = []

    def check(self, samples: List) -> List:
        """

        Parameters
        ----------
        samples: list
            names or numbers, or pymatgen.structures

        Returns
        -------
        result: list
            list of filtered structure.
        """

        self.mark = []
        structures_t = []
        for i, si in enumerate(samples):
            si_ = self.func(si)
            if np.all([True if ei in self.check_method else False for ei in si_]):
                structures_t.append(si)
                self.mark.append(1)
            else:
                if hasattr(si, "composition"):
                    print("The {} (st,ed,th) sample {}".format(i, str(si.composition)),
                          "is with element out of AVAILABLE_ELE_NAME\n"
                          "please to check_data.py for more information.")
                elif hasattr(si, "get_chemical_formula"):
                    print("The {} (st,ed,th) sample {}".format(i, str(si.get_chemical_formula())),
                          "is with element out of AVAILABLE_ELE_NAME\n"
                          "please to check_data.py for more information.")
                else:
                    print("The {} (st,ed,th) sample {}".format(i, str(si)),
                          "is with element out of AVAILABLE_ELE_NAME\n"
                          "please to check_data.py for more information.")
                self.mark.append(0)
        return structures_t

    def passed_idx(self) -> np.ndarray:
        """The mark for all structures, return np.ndarray index"""
        return np.where(np.array(self.mark) == 1)[0]

    @classmethod
    def from_list(cls, check_method="name", grouped="False"):
        """Get checker for list"""
        if grouped:
            func = lambda x: x
        else:
            func = lambda x: [x, ]
        return cls(check_method=check_method, func=func)

    @classmethod
    def from_pymatgen_structures(cls):
        """Get checker for pymatgen.Structure"""
        func = lambda x: x.atomic_numbers
        return cls(check_method="number", func=func)
